# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from gammapy.catalog import SourceCatalog1LHAASO
from gammapy.modeling.models import (
    GaussianSpatialModel,
    PointSpatialModel,
    PowerLawNormSpectralModel,
    PowerLawSpectralModel,
    TemplateSpatialModel,
)
from gammapy.utils.testing import requires_data


@pytest.fixture(scope="session")
def lhaaso1():
    return SourceCatalog1LHAASO()


@requires_data()
class TestSourceCatalog1LHAASO:
    @staticmethod
    def test_source_table(lhaaso1):
        assert lhaaso1.tag == "1LHAASO"
        assert len(lhaaso1.table) == 90

    @staticmethod
    def test_positions(lhaaso1):
        assert len(lhaaso1.positions) == 90

    @staticmethod
    def test_to_models(lhaaso1):
        models = lhaaso1.to_models(which="both")
        assert len(models) == 90

        models = lhaaso1.to_models(which="KM2A")
        assert np.all(
            [m.spectral_model.reference.quantity == 50 * u.TeV for m in models]
        )
        assert len(models) == 75
        models = lhaaso1.to_models(which="WCDA")
        assert np.all(
            [m.spectral_model.reference.quantity == 3 * u.TeV for m in models]
        )
        assert len(models) == 69


@requires_data()
class TestSourceCatalogObject1LHAASO:
    @staticmethod
    def test_data(lhaaso1):
        assert lhaaso1[0].data["Source_Name"] == "1LHAASO J0007+5659u"
        assert "KM2A" in lhaaso1[0].data["Model_a"]

        assert_allclose(lhaaso1[0].data["r39_ul"].value, 0.18)
        assert lhaaso1[0].data["r39_ul"].unit == u.deg

        assert_allclose(lhaaso1[0].data["N0"].value, 0.33e-16)
        assert lhaaso1[0].data["N0"].unit == u.Unit("cm−2 s−1 TeV−1")

        assert_allclose(lhaaso1[0].data["N0_err"].value, 0.05e-16)
        assert lhaaso1[0].data["N0_err"].unit == u.Unit("cm−2 s−1 TeV−1")

        assert_allclose(lhaaso1[0].data["N0_ul_b"].value, 0.27e-13)
        assert lhaaso1[0].data["N0_ul_b"].unit == u.Unit("cm−2 s−1 TeV−1")

        assert lhaaso1[1].data["ASSO_Name"] == "CTA 1"
        assert_allclose(lhaaso1[1].data["ASSO_Sep"].value, 0.12)
        assert lhaaso1[0].data["ASSO_Sep"].unit == u.deg

        assert lhaaso1[10].data["Source_Name"] == "1LHAASO J0428+5531*"
        assert "WCDA" in lhaaso1[10].data["Model_a"]

        assert_allclose(lhaaso1[10].data["RAJ2000"].value, 67.23)
        assert_allclose(lhaaso1[10].data["DECJ2000"].value, 55.53)
        assert_allclose(lhaaso1[10].data["pos_err"].value, 0.36)

        assert lhaaso1[10].data["RAJ2000"].unit == u.deg
        assert lhaaso1[10].data["DECJ2000"].unit == u.deg
        assert lhaaso1[10].data["pos_err"].unit == u.deg

        assert_allclose(lhaaso1[10].data["r39"].value, 1.18)
        assert_allclose(lhaaso1[10].data["r39_b"].value, 0.32)
        assert lhaaso1[10].data["r39_b"].unit == u.deg

        assert_allclose(lhaaso1[10].data["r39_err"].value, 0.12)
        assert_allclose(lhaaso1[10].data["r39_err_b"].value, 0.06)
        assert lhaaso1[10].data["r39_err_b"].unit == u.deg

    @staticmethod
    def test_position(lhaaso1):
        position = lhaaso1[0].position
        assert_allclose(position.ra.deg, 1.86, atol=1e-3)
        assert_allclose(position.dec.deg, 57.00, atol=1e-3)

    @staticmethod
    def test_sky_model(lhaaso1):
        model = lhaaso1[0].sky_model("both")
        assert model.name == "1LHAASO J0007+5659u"
        assert isinstance(model.spectral_model, PowerLawSpectralModel)
        assert isinstance(model.spatial_model, PointSpatialModel)

        assert lhaaso1[0].sky_model("WCDA") is None

        model = lhaaso1[1].sky_model("both")
        assert model.name == "1LHAASO J0007+7303u"
        assert isinstance(model.spectral_model, PowerLawNormSpectralModel)
        assert isinstance(model.spatial_model, TemplateSpatialModel)

        model = lhaaso1[1].sky_model("KM2A")
        assert model.name == "1LHAASO J0007+7303u"
        assert isinstance(model.spectral_model, PowerLawSpectralModel)
        assert isinstance(model.spatial_model, GaussianSpatialModel)

        model = lhaaso1[1].sky_model("WCDA")
        assert model.name == "1LHAASO J0007+7303u"
        assert isinstance(model.spectral_model, PowerLawSpectralModel)
        assert isinstance(model.spatial_model, PointSpatialModel)

        model = lhaaso1[11].sky_model("both")
        assert model.name == "1LHAASO J0500+4454"
        assert isinstance(model.spectral_model, PowerLawSpectralModel)
        assert isinstance(model.spatial_model, GaussianSpatialModel)

    @staticmethod
    def test_spectral_model(lhaaso1):
        m = lhaaso1[0].spectral_model("KM2A")
        dnde, dnde_errn, dnde_errp = m.evaluate_error(50 * u.TeV)
        assert dnde.unit == "cm-2 s-1 TeV-1"
        assert_allclose(dnde.value, 0.33e-16, rtol=5e-2)
        assert_allclose(dnde_errn.value, 0.05e-16, rtol=5e-2)
        assert_allclose(dnde_errp.value, 0.05e-16, rtol=5e-2)

        m = lhaaso1[11].spectral_model("WCDA")
        dnde, dnde_errn, dnde_errp = m.evaluate_error(3 * u.TeV)
        assert dnde.unit == "cm-2 s-1 TeV-1"
        assert_allclose(dnde.value, 0.69e-13, rtol=5e-2)
        assert_allclose(dnde_errn.value, 0.16e-13, rtol=5e-2)
        assert_allclose(dnde_errp.value, 0.16e-13, rtol=5e-2)

    @staticmethod
    def test_spatial_model(lhaaso1):
        m = lhaaso1[0].spatial_model("KM2A")

        assert isinstance(m, PointSpatialModel)
        assert m.lon_0.unit == "deg"
        assert_allclose(m.lon_0.value, 1.86, atol=1e-2)
        assert_allclose(m.lon_0.error, 0.09, atol=1e-2)
        assert m.lat_0.unit == "deg"
        assert_allclose(m.lat_0.value, 57.00, atol=1e-2)
        assert_allclose(m.lat_0.error, 0.049, atol=1e-2)
        assert m.frame == "fk5"

        m = lhaaso1[11].spatial_model("WCDA")
        assert isinstance(m, GaussianSpatialModel)
        assert m.lon_0.unit == "deg"
        assert_allclose(m.lon_0.value, 75.01, atol=1e-10)
        assert m.lat_0.unit == "deg"
        assert_allclose(m.lat_0.value, 44.92, atol=1e-10)
        assert m.frame == "fk5"
        assert m.sigma.unit == "deg"
        assert_allclose(m.sigma.value, 0.41, atol=1e-3)

        model = lhaaso1["1LHAASO J0007+5659u"].spatial_model("KM2A")
        pos_err = model.position_error
        assert_allclose(pos_err.height.value, 2 * 0.12, rtol=1e-4)
        assert_allclose(pos_err.width.value, 2 * 0.12, rtol=1e-4)
        assert_allclose(model.position.ra.value, pos_err.center.ra.value)
        assert_allclose(model.position.dec.value, pos_err.center.dec.value)



from numpy.testing import assert_allclose
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from gammapy.datasets import MapDataset
from gammapy.estimators.energydependentmorphology import (
    EnergyDependentMorphologyEstimator,
    weighted_chi2_parameter,
)
from gammapy.modeling.models import (
    GaussianSpatialModel,
    PowerLawSpectralModel,
    SkyModel,
)
from gammapy.utils.testing import requires_data


@requires_data()
class TestEnergyDependentEstimator:
    def setup_class(self):
        stacked_dataset = MapDataset.read(
            "$GAMMAPY_DATA/estimators/mock_DL4/dataset_energy_dependent.fits.gz"
        )
        energy_edges = [1, 5, 20] * u.TeV

        source_pos = SkyCoord(5.58, 0.2, unit="deg", frame="galactic")

        spectral_model = PowerLawSpectralModel(
            index=2.94,
            amplitude=9.8e-12 * u.Unit("cm-2 s-1 TeV-1"),
            reference=1.0 * u.TeV,
        )
        spatial_model = GaussianSpatialModel(
            lon_0=source_pos.l, lat_0=source_pos.b, frame="galactic", sigma=0.2 * u.deg
        )

        model = SkyModel(
            spatial_model=spatial_model, spectral_model=spectral_model, name="source"
        )

        model.spectral_model.index.frozen = True

        model.spatial_model.lon_0.min = source_pos.galactic.l.deg - 0.8
        model.spatial_model.lon_0.max = source_pos.galactic.l.deg + 0.8
        model.spatial_model.lat_0.min = source_pos.galactic.b.deg - 0.8
        model.spatial_model.lat_0.max = source_pos.galactic.b.deg + 0.8

        stacked_dataset.models = model
        estimator = EnergyDependentMorphologyEstimator(
            energy_edges=energy_edges, source="source"
        )
        self.estimator_result = estimator.run(stacked_dataset)

    def test_edep(self):
        results_edep = self.estimator_result["energy_dependence"]["result"]
        assert_allclose(
            results_edep["lon_0"],
            [5.606467, 5.608664, 5.597394] * u.deg,
            rtol=1e-3,
        )
        assert_allclose(
            results_edep["lat_0"],
            [0.20289353, 0.20589559, 0.18106776] * u.deg,
            rtol=1e-3,
        )
        assert_allclose(
            results_edep["sigma"],
            [0.21709024, 0.2315993, 0.13505759] * u.deg,
            rtol=1e-3,
        )
        assert_allclose(
            self.estimator_result["energy_dependence"]["delta_ts"], 50.719, rtol=1e-3
        )

    def test_significance(self):
        results_src = self.estimator_result["src_above_bkg"]
        assert_allclose(
            results_src["delta_ts"],
            [1683.1128, 289.8156],
            rtol=1e-3,
        )
        assert_allclose(
            results_src["significance"],
            [np.inf, 16.546],
            rtol=1e-3,
        )

    def test_chi2(self):
        results_edep = self.estimator_result["energy_dependence"]["result"]
        chi2_sigma = weighted_chi2_parameter(
            results_edep, parameters=["sigma", "lat_0", "lon_0"]
        )

        assert_allclose(
            chi2_sigma["chi2"],
            [75.735899, 1.962729, 0.413059],
            rtol=1e-2,
        )

        assert_allclose(
            chi2_sigma["significance"],
            [8.7026, 1.4009, 0.64269],
            rtol=1e-2,
        )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Dark matter profiles."""

import abc
import html
import numpy as np
import astropy.units as u
from gammapy.modeling import Parameter, Parameters
from gammapy.utils.integrate import trapz_loglog

__all__ = [
    "BurkertProfile",
    "DMProfile",
    "EinastoProfile",
    "IsothermalProfile",
    "MooreProfile",
    "NFWProfile",
    "ZhaoProfile",
]


class DMProfile(abc.ABC):
    """DMProfile model base class."""

    LOCAL_DENSITY = 0.3 * u.GeV / (u.cm**3)
    """Local dark matter density as given in reference 2"""
    DISTANCE_GC = 8.33 * u.kpc
    """Distance to the Galactic Center as given in reference 2"""

    def __call__(self, radius):
        """Call evaluate method of derived classes."""
        kwargs = {par.name: par.quantity for par in self.parameters}
        return self.evaluate(radius, **kwargs)

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def scale_to_local_density(self):
        """Scale to local density."""
        scale = (self.LOCAL_DENSITY / self(self.DISTANCE_GC)).to_value("")
        self.parameters["rho_s"].value *= scale

    def _eval_substitution(self, radius, separation, squared):
        """Density at given radius together with the substitution part."""
        exponent = 2 if squared else 1
        return (
            self(radius) ** exponent
            * radius
            / np.sqrt(radius**2 - (self.DISTANCE_GC * np.sin(separation)) ** 2)
        )

    def integral(self, rmin, rmax, separation, ndecade, squared=True):
        r"""Integrate dark matter profile numerically.

        .. math::
            F(r_{min}, r_{max}) = \int_{r_{min}}^{r_{max}}\rho(r)^\gamma dr \\
            \gamma = 2 \text{for annihilation} \\
            \gamma = 1 \text{for decay}

        Parameters
        ----------
        rmin, rmax : `~astropy.units.Quantity`
            Lower and upper bound of integration range.
        separation : `~numpy.ndarray`
            Separation angle in radians.
        ndecade : int, optional
            Number of grid points per decade used for the integration.
            Default is 10000.
        squared : bool, optional
            Square the profile before integration.
            Default is True.
        """
        integral = self.integrate_spectrum_separation(
            self._eval_substitution, rmin, rmax, separation, ndecade, squared
        )
        inegral_unit = u.Unit("GeV2 cm-5") if squared else u.Unit("GeV cm-2")
        return integral.to(inegral_unit)

    def integrate_spectrum_separation(
        self, func, xmin, xmax, separation, ndecade, squared=True
    ):
        """Squared dark matter profile integral.

        Parameters
        ----------
        xmin, xmax : `~astropy.units.Quantity`
            Lower and upper bound of integration range.
        separation : `~numpy.ndarray`
            Separation angle in radians.
        ndecade : int
            Number of grid points per decade used for the integration.
        squared : bool
            Square the profile before integration.
            Default is True.
        """
        unit = xmin.unit
        xmin = xmin.value
        xmax = xmax.to_value(unit)
        logmin = np.log10(xmin)
        logmax = np.log10(xmax)
        n = np.int32((logmax - logmin) * ndecade)
        x = np.logspace(logmin, logmax, n) * unit
        y = func(x, separation, squared)
        val = trapz_loglog(y, x)
        return val.sum()


class ZhaoProfile(DMProfile):
    r"""Zhao Profile.

    It is a generalization of the NFW profile. The volume density
    is parametrized with a double power-law. Scale radii smaller than the scale radius are described with a slope of
    :math:`-\gamma` and scale radii larger than the scale radius are described with a slope of :math:`-\beta`.
    :math:`\alpha` is a measure for the width of the transition region.

    .. math::
        \rho(r) = \rho_s \left(\frac{r_s}{r}\right)^\gamma \left(1 + \left(\frac{r}{r_s}\right)^\frac{1}{\alpha} \right)^{(\gamma - \beta) \alpha}

    Equation (1) from [1]_.

    Parameters
    ----------
    r_s : `~astropy.units.Quantity`
        Scale radius, :math:`r_s`.
    alpha : `~astropy.units.Quantity`
        :math:`\alpha`.
    beta: `~astropy.units.Quantity`
        :math:`\beta`.
    gamma : `~astropy.units.Quantity`
        :math:`\gamma`.
    rho_s : `~astropy.units.Quantity`
        Characteristic density, :math:`\rho_s`.

    References
    ----------
    .. [1] `Zhao (1996), "Analytical models for galactic nuclei"
      <https://ui.adsabs.harvard.edu/abs/1996MNRAS.278..488Z>`_
    * `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection"
      <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    """

    DEFAULT_SCALE_RADIUS = 24.42 * u.kpc
    DEFAULT_ALPHA = 1
    DEFAULT_BETA = 3
    DEFAULT_GAMMA = 1
    """
    (alpha, beta, gamma) = (1,3,1) is NFW profile.
    Default scale radius as given in reference 2 (same as for NFW profile)
    """

    def __init__(
        self, r_s=None, alpha=None, beta=None, gamma=None, rho_s=1 * u.Unit("GeV / cm3")
    ):
        r_s = self.DEFAULT_SCALE_RADIUS if r_s is None else r_s
        alpha = self.DEFAULT_ALPHA if alpha is None else alpha
        beta = self.DEFAULT_BETA if beta is None else beta
        gamma = self.DEFAULT_GAMMA if gamma is None else gamma
        self.parameters = Parameters(
            [
                Parameter("r_s", u.Quantity(r_s)),
                Parameter("rho_s", u.Quantity(rho_s)),
                Parameter("alpha", alpha),
                Parameter("beta", beta),
                Parameter("gamma", gamma),
            ]
        )

    @staticmethod
    def evaluate(radius, r_s, alpha, beta, gamma, rho_s):
        """Evaluate the profile."""
        rr = radius / r_s
        return rho_s / (rr**gamma * (1 + rr ** (1 / alpha)) ** ((beta - gamma) * alpha))


class NFWProfile(DMProfile):
    r"""NFW Profile.

    .. math::
        \rho(r) = \rho_s \frac{r_s}{r}\left(1 + \frac{r}{r_s}\right)^{-2}

    Parameters
    ----------
    r_s : `~astropy.units.Quantity`
        Scale radius, :math:`r_s`.
    rho_s : `~astropy.units.Quantity`
        Characteristic density, :math:`\rho_s`.

    References
    ----------
    * `Navarro et al. (1997), "A Universal Density Profile from Hierarchical Clustering"
      <https://ui.adsabs.harvard.edu/abs/1997ApJ...490..493N>`_
    * `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection"
      <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    """

    DEFAULT_SCALE_RADIUS = 24.42 * u.kpc
    """Default scale radius as given in reference 2"""

    def __init__(self, r_s=None, rho_s=1 * u.Unit("GeV / cm3")):
        r_s = self.DEFAULT_SCALE_RADIUS if r_s is None else r_s
        self.parameters = Parameters(
            [Parameter("r_s", u.Quantity(r_s)), Parameter("rho_s", u.Quantity(rho_s))]
        )

    @staticmethod
    def evaluate(radius, r_s, rho_s):
        """Evaluate the profile."""
        rr = radius / r_s
        return rho_s / (rr * (1 + rr) ** 2)


class EinastoProfile(DMProfile):
    r"""Einasto Profile.

    .. math::
        \rho(r) = \rho_s \exp{
            \left(-\frac{2}{\alpha}\left[
            \left(\frac{r}{r_s}\right)^{\alpha} - 1\right] \right)}

    Parameters
    ----------
    r_s : `~astropy.units.Quantity`
        Scale radius, :math:`r_s`.
    alpha : `~astropy.units.Quantity`
        :math:`\alpha`.
    rho_s : `~astropy.units.Quantity`
        Characteristic density, :math:`\rho_s`.

    References
    ----------
    * `Einasto (1965), "On the Construction of a Composite Model for the Galaxy and on the Determination of the System
      of Galactic Parameters" <https://ui.adsabs.harvard.edu/abs/1965TrAlm...5...87E>`_
    * `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection"
      <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    """

    DEFAULT_SCALE_RADIUS = 28.44 * u.kpc
    """Default scale radius as given in reference 2"""
    DEFAULT_ALPHA = 0.17
    """Default scale radius as given in reference 2"""

    def __init__(self, r_s=None, alpha=None, rho_s=1 * u.Unit("GeV / cm3")):
        alpha = self.DEFAULT_ALPHA if alpha is None else alpha
        r_s = self.DEFAULT_SCALE_RADIUS if r_s is None else r_s

        self.parameters = Parameters(
            [
                Parameter("r_s", u.Quantity(r_s)),
                Parameter("alpha", u.Quantity(alpha)),
                Parameter("rho_s", u.Quantity(rho_s)),
            ]
        )

    @staticmethod
    def evaluate(radius, r_s, alpha, rho_s):
        """Evaluate the profile."""
        rr = radius / r_s
        exponent = (2 / alpha) * (rr**alpha - 1)
        return rho_s * np.exp(-1 * exponent)


class IsothermalProfile(DMProfile):
    r"""Isothermal Profile.

    .. math:: \rho(r) = \frac{\rho_s}{1 + (r/r_s)^2}

    Parameters
    ----------
    r_s : `~astropy.units.Quantity`
        Scale radius, :math:`r_s`.

    References
    ----------
    * `Begeman et al. (1991), "Extended rotation curves of spiral galaxies : dark haloes and modified dynamics"
      <https://ui.adsabs.harvard.edu/abs/1965TrAlm...5...87E>`_
    * `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection"
      <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    """

    DEFAULT_SCALE_RADIUS = 4.38 * u.kpc
    """Default scale radius as given in reference 2"""

    def __init__(self, r_s=None, rho_s=1 * u.Unit("GeV / cm3")):
        r_s = self.DEFAULT_SCALE_RADIUS if r_s is None else r_s

        self.parameters = Parameters(
            [Parameter("r_s", u.Quantity(r_s)), Parameter("rho_s", u.Quantity(rho_s))]
        )

    @staticmethod
    def evaluate(radius, r_s, rho_s):
        """Evaluate the profile."""
        rr = radius / r_s
        return rho_s / (1 + rr**2)


class BurkertProfile(DMProfile):
    r"""Burkert Profile.

    .. math:: \rho(r) = \frac{\rho_s}{(1 + r/r_s)(1 + (r/r_s)^2)}

    Parameters
    ----------
    r_s : `~astropy.units.Quantity`
        Scale radius, :math:`r_s`.

    References
    ----------
    * `Burkert (1995), "The Structure of Dark Matter Halos in Dwarf Galaxies"
      <https://ui.adsabs.harvard.edu/abs/1965TrAlm...5...87E>`_
    * `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection"
      <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    """

    DEFAULT_SCALE_RADIUS = 12.67 * u.kpc
    """Default scale radius as given in reference 2"""

    def __init__(self, r_s=None, rho_s=1 * u.Unit("GeV / cm3")):
        r_s = self.DEFAULT_SCALE_RADIUS if r_s is None else r_s

        self.parameters = Parameters(
            [Parameter("r_s", u.Quantity(r_s)), Parameter("rho_s", u.Quantity(rho_s))]
        )

    @staticmethod
    def evaluate(radius, r_s, rho_s):
        """Evaluate the profile."""
        rr = radius / r_s
        return rho_s / ((1 + rr) * (1 + rr**2))


class MooreProfile(DMProfile):
    r"""Moore Profile.

    .. math::
        \rho(r) = \rho_s \left(\frac{r_s}{r}\right)^{1.16}
        \left(1 + \frac{r}{r_s} \right)^{-1.84}

    Parameters
    ----------
    r_s : `~astropy.units.Quantity`
        Scale radius, :math:`r_s`.

    References
    ----------
    * `Diemand et al. (2004), "Convergence and scatter of cluster density profiles"
      <https://ui.adsabs.harvard.edu/abs/1965TrAlm...5...87E>`_
    * `Marco et al. (2011), "PPPC 4 DM ID: a poor particle physicist cookbook for dark matter indirect detection"
      <https://ui.adsabs.harvard.edu/abs/2011JCAP...03..051C>`_
    """

    DEFAULT_SCALE_RADIUS = 30.28 * u.kpc
    """Default scale radius as given in reference 2"""

    def __init__(self, r_s=None, rho_s=1 * u.Unit("GeV / cm3")):
        r_s = self.DEFAULT_SCALE_RADIUS if r_s is None else r_s

        self.parameters = Parameters(
            [Parameter("r_s", u.Quantity(r_s)), Parameter("rho_s", u.Quantity(rho_s))]
        )

    @staticmethod
    def evaluate(radius, r_s, rho_s):
        """Evaluate the profile."""
        rr = radius / r_s
        rr_ = r_s / radius
        return rho_s * rr_**1.16 * (1 + rr) ** (-1.84)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from itertools import repeat
from gammapy.utils import parallel as parallel
from gammapy.utils.scripts import make_path


class ObservationsEventsSampler(parallel.ParallelMixin):
    """Run event sampling for an ensemble of observations.

    Parameters
    ----------
    sampler_kwargs : dict, optional
        Arguments passed to `~gammapy.datasets.MapDatasetEventSampler`.
    dataset_kwargs : dict, optional
        Arguments passed to `~gammapy.datasets.create_map_dataset_from_observation()`.
    outdir : str, optional
        Path of the output files created. Default is "./simulated_data/".
        If None, a list of `~gammapy.data.Observation` is returned.
    overwrite : bool, optional
        Overwrite existing file. Default is True.
    n_jobs : int, optional
        Number of processes to run in parallel.
        By default, the value is 1, unless `~gammapy.utils.parallel.N_JOBS_DEFAULT` has been modified.
    parallel_backend : {'multiprocessing', 'ray'}, optional
        Which backend to use for multiprocessing.
        Default is None.
    """

    def __init__(
        self,
        sampler_kwargs=None,
        dataset_kwargs=None,
        outdir="./simulated_data/",
        overwrite=True,
        n_jobs=None,
        parallel_backend=None,
    ):
        if outdir is not None:
            outdir = make_path(outdir)
            outdir.mkdir(exist_ok=True, parents=True)
        self.outdir = outdir
        self.n_jobs = n_jobs
        self.parallel_backend = parallel_backend
        self.overwrite = overwrite

        if sampler_kwargs is None:
            sampler_kwargs = {}
        self.sampler_kwargs = sampler_kwargs
        self.dataset_kwargs = dataset_kwargs

    def simulate_observation(self, observation, models=None):
        """Simulate a single observation.

        Parameters
        ----------
        observation : `~gammapy.data.Observation`
            Observation to be simulated.
        models : `~gammapy.modeling.models.Models`, optional
            Models to simulate.
            Can be None to only sample background events. Default is None.
        """
        from gammapy.datasets import ObservationEventSampler

        sampler = ObservationEventSampler(
            **self.sampler_kwargs, dataset_kwargs=self.dataset_kwargs
        )
        observation = sampler.run(observation, models=models)

        if self.outdir is not None:
            observation.write(
                self.outdir / f"obs_{observation.obs_id}.fits",
                overwrite=self.overwrite,
            )
        else:
            return observation

    def run(self, observations, models=None):
        """Run event sampling for an ensemble of observations.

        Parameters
        ----------
        observations : `~gammapy.data.Observations`
            Observations to be simulated.
        models : `~gammapy.modeling.models.Models`, optional
            Models to simulate.
            Can be None to only sample background events. Default is None.
        """
        n_jobs = min(self.n_jobs, len(observations))

        observations = parallel.run_multiprocessing(
            self.simulate_observation,
            zip(
                observations,
                repeat(models),
            ),
            backend=self.parallel_backend,
            pool_kwargs=dict(processes=n_jobs),
            task_name="Simulate observations",
        )
        if self.outdir is None:
            return observations



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import html
import json
import logging
from collections import defaultdict
from collections.abc import Mapping
from enum import Enum
from pathlib import Path
from typing import List, Optional
from astropy import units as u
import yaml
from pydantic import BaseModel, ConfigDict
from gammapy.makers import MapDatasetMaker
from gammapy.utils.scripts import read_yaml, to_yaml, write_yaml
from gammapy.utils.types import AngleType, EnergyType, PathType, TimeType

__all__ = ["WorkflowConfig"]

CONFIG_PATH = Path(__file__).resolve().parent / "config"
DOCS_FILE = CONFIG_PATH / "docs.yaml"

log = logging.getLogger(__name__)


def deep_update(d, u):
    """Recursively update a nested dictionary.

    Taken from: https://stackoverflow.com/a/3233356/19802442
    """
    for k, v in u.items():
        if isinstance(v, Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class WorkflowStepEnum(str, Enum):
    data_reduction = "data-selection"
    observations = "observations"
    datasets = "datasets"
    excess_map = "excess-map"
    fit = "fit"
    flux_points = "flux-points"
    light_curve = "light-curve"


class ReductionTypeEnum(str, Enum):
    spectrum = "1d"
    cube = "3d"


class FrameEnum(str, Enum):
    icrs = "icrs"
    galactic = "galactic"


class RequiredHDUEnum(str, Enum):
    events = "events"
    gti = "gti"
    aeff = "aeff"
    bkg = "bkg"
    edisp = "edisp"
    psf = "psf"
    rad_max = "rad_max"


class BackgroundMethodEnum(str, Enum):
    reflected = "reflected"
    fov = "fov_background"
    ring = "ring"


class SafeMaskMethodsEnum(str, Enum):
    aeff_default = "aeff-default"
    aeff_max = "aeff-max"
    edisp_bias = "edisp-bias"
    offset_max = "offset-max"
    bkg_peak = "bkg-peak"


class MapSelectionEnum(str, Enum):
    counts = "counts"
    exposure = "exposure"
    background = "background"
    psf = "psf"
    edisp = "edisp"


class GammapyBaseConfig(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True,
        validate_assignment=True,
        extra="forbid",
        validate_default=True,
        use_enum_values=True,
        json_encoders={u.Quantity: lambda v: f"{v.value} {v.unit}"},
    )

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"


class SkyCoordConfig(GammapyBaseConfig):
    frame: Optional[FrameEnum] = None
    lon: Optional[AngleType] = None
    lat: Optional[AngleType] = None


class EnergyAxisConfig(GammapyBaseConfig):
    min: Optional[EnergyType] = None
    max: Optional[EnergyType] = None
    nbins: Optional[int] = None


class SpatialCircleConfig(GammapyBaseConfig):
    frame: Optional[FrameEnum] = None
    lon: Optional[AngleType] = None
    lat: Optional[AngleType] = None
    radius: Optional[AngleType] = None


class EnergyRangeConfig(GammapyBaseConfig):
    min: Optional[EnergyType] = None
    max: Optional[EnergyType] = None


class TimeRangeConfig(GammapyBaseConfig):
    start: Optional[TimeType] = None
    stop: Optional[TimeType] = None


class FluxPointsConfig(GammapyBaseConfig):
    energy: EnergyAxisConfig = EnergyAxisConfig()
    source: str = "source"
    parameters: dict = {"selection_optional": "all"}


class LightCurveConfig(GammapyBaseConfig):
    time_intervals: TimeRangeConfig = TimeRangeConfig()
    energy_edges: EnergyAxisConfig = EnergyAxisConfig()
    source: str = "source"
    parameters: dict = {"selection_optional": "all"}


class FitConfig(GammapyBaseConfig):
    fit_range: EnergyRangeConfig = EnergyRangeConfig()


class ExcessMapConfig(GammapyBaseConfig):
    correlation_radius: AngleType = "0.1 deg"
    parameters: dict = {}
    energy_edges: EnergyAxisConfig = EnergyAxisConfig()


class BackgroundConfig(GammapyBaseConfig):
    method: Optional[BackgroundMethodEnum] = None
    exclusion: Optional[PathType] = None
    parameters: dict = {}


class SafeMaskConfig(GammapyBaseConfig):
    methods: List[SafeMaskMethodsEnum] = [SafeMaskMethodsEnum.aeff_default]
    parameters: dict = {}


class EnergyAxesConfig(GammapyBaseConfig):
    energy: EnergyAxisConfig = EnergyAxisConfig(min="1 TeV", max="10 TeV", nbins=5)
    energy_true: EnergyAxisConfig = EnergyAxisConfig(
        min="0.5 TeV", max="20 TeV", nbins=16
    )


class SelectionConfig(GammapyBaseConfig):
    offset_max: AngleType = "2.5 deg"


class WidthConfig(GammapyBaseConfig):
    width: AngleType = "5 deg"
    height: AngleType = "5 deg"


class WcsConfig(GammapyBaseConfig):
    skydir: SkyCoordConfig = SkyCoordConfig()
    binsize: AngleType = "0.02 deg"
    width: WidthConfig = WidthConfig()
    binsize_irf: AngleType = "0.2 deg"


class GeomConfig(GammapyBaseConfig):
    wcs: WcsConfig = WcsConfig()
    selection: SelectionConfig = SelectionConfig()
    axes: EnergyAxesConfig = EnergyAxesConfig()


class DatasetsConfig(GammapyBaseConfig):
    type: ReductionTypeEnum = ReductionTypeEnum.spectrum
    stack: bool = True
    geom: GeomConfig = GeomConfig()
    map_selection: List[MapSelectionEnum] = MapDatasetMaker.available_selection
    background: BackgroundConfig = BackgroundConfig()
    safe_mask: SafeMaskConfig = SafeMaskConfig()
    on_region: SpatialCircleConfig = SpatialCircleConfig()
    containment_correction: bool = True


class ObservationsConfig(GammapyBaseConfig):
    datastore: PathType = Path("$GAMMAPY_DATA/hess-dl3-dr1/")
    obs_ids: List[int] = []
    obs_file: Optional[PathType] = None
    obs_cone: SpatialCircleConfig = SpatialCircleConfig()
    obs_time: TimeRangeConfig = TimeRangeConfig()
    required_irf: List[RequiredHDUEnum] = ["aeff", "edisp", "psf", "bkg"]


class LogConfig(GammapyBaseConfig):
    level: str = "info"
    filename: Optional[PathType] = None
    filemode: Optional[str] = None
    format: Optional[str] = None
    datefmt: Optional[str] = None


class GeneralConfig(GammapyBaseConfig):
    log: LogConfig = LogConfig()
    outdir: str = "."
    n_jobs: int = 1
    datasets_file: Optional[PathType] = None
    models_file: Optional[PathType] = None
    steps: List[str] = ["data-selection", "fit", "flux-points"]
    overwrite: List[bool] = [True, True, True]


class WorkflowConfig(GammapyBaseConfig):
    """Gammapy workflow configuration."""

    general: GeneralConfig = GeneralConfig()
    observations: ObservationsConfig = ObservationsConfig()
    datasets: DatasetsConfig = DatasetsConfig()
    fit: FitConfig = FitConfig()
    flux_points: FluxPointsConfig = FluxPointsConfig()
    excess_map: ExcessMapConfig = ExcessMapConfig()
    light_curve: LightCurveConfig = LightCurveConfig()

    def __str__(self):
        """Display settings in pretty YAML format."""
        info = self.__class__.__name__ + "\n\n\t"
        data = self.to_yaml()
        data = data.replace("\n", "\n\t")
        info += data
        return info.expandtabs(tabsize=4)

    @classmethod
    def read(cls, path):
        """Read from YAML file.

        Parameters
        ----------
        path : str
            input filepath
        """
        config = read_yaml(path)
        config.pop("metadata", None)
        return WorkflowConfig(**config)

    @classmethod
    def from_yaml(cls, config_str):
        """Create from YAML string.

        Parameters
        ----------
        config_str : str
            yaml str

        """
        settings = yaml.safe_load(config_str)
        return WorkflowConfig(**settings)

    def write(self, path, overwrite=False):
        """Write to YAML file.

        Parameters
        ----------
        path : `pathlib.Path` or str
            Path to write files.
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        """
        yaml_str = self.to_yaml()
        write_yaml(yaml_str, path, overwrite=overwrite)

    def to_yaml(self):
        """Convert to YAML string."""
        data = json.loads(self.model_dump_json())
        return to_yaml(data)

    def set_logging(self):
        """Set logging config.

        Calls ``logging.basicConfig``, i.e. adjusts global logging state.
        """
        self.general.log.level = self.general.log.level.upper()
        logging.basicConfig(**self.general.log.model_dump())
        log.info("Setting logging config: {!r}".format(self.general.log.model_dump()))

    def update(self, config=None):
        """Update config with provided settings.

        Parameters
        ----------
        config : str or `WorkflowConfig` object, optional
            Configuration settings provided in dict() syntax. Default is None.
        """
        if isinstance(config, str):
            other = WorkflowConfig.from_yaml(config)
        elif isinstance(config, WorkflowConfig):
            other = config
        else:
            raise TypeError(f"Invalid type: {config}")

        config_new = deep_update(
            self.model_dump(exclude_defaults=True),
            other.model_dump(exclude_defaults=True),
        )
        return WorkflowConfig(**config_new)

    @staticmethod
    def _get_doc_sections():
        """Return dictionary with commented docs from docs file."""
        doc = defaultdict(str)
        with open(DOCS_FILE) as f:
            for line in filter(lambda line: not line.startswith("---"), f):
                line = line.strip("\n")
                if line.startswith("# Section: "):
                    keyword = line.replace("# Section: ", "")
                doc[keyword] += line + "\n"
        return doc



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Gammapy documentation generator utilities.

This is for the Sphinx docs build of Gammapy itself,
it is used from ``docs/conf.py``.

It should not be imported or used from other scripts or packages,
because we will change it for our use case whenever we like.

Docutils / Sphinx is notoriously hard to understand / learn about.
Here's some good resources with working examples:

- https://doughellmann.com/blog/2010/05/09/defining-custom-roles-in-sphinx/
- http://docutils.sourceforge.net/docs/howto/rst-directives.html
- https://github.com/docutils/docutils/blob/master/docutils/docutils/parsers/rst/directives/images.py
- https://github.com/sphinx-doc/sphinx/blob/master/sphinx/directives/other.py
- https://github.com/bokeh/bokeh/tree/master/src/bokeh/sphinxext
"""

import os
import re
from pathlib import Path
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import register_directive
from docutils.parsers.rst.directives.body import CodeBlock
from docutils.parsers.rst.directives.images import Image
from docutils.parsers.rst.directives.misc import Include, Raw
from docutils import nodes
from sphinx.util import logging
from sphinx.transforms.post_transforms import SphinxPostTransform
from gammapy.analysis import AnalysisConfig

try:
    gammapy_data_path = Path(os.environ["GAMMAPY_DATA"])
    HAS_GP_DATA = True
except KeyError:
    HAS_GP_DATA = False

log = logging.getLogger(__name__)


class AccordionHeader(Directive):
    """Insert HTML code to open an accordion box in the How To."""

    option_spec = {"id": str, "title": str, "link": str}

    def run(self):
        raw = f"""
            <div id="accordion" class="shadow tutorial-accordion">
        <div class="card tutorial-card">
            <div class="card-header d-flex justify-content-between align-items-center">
                <div class="d-flex align-items-center" data-bs-toggle="collapse"
                data-bs-target="#{self.options["id"]}" style="cursor: pointer;">
                    <button class="btn btn-dark btn-sm me-2"
                    style="background-color: #150458; border-radius: 50%; width: 24px; height: 24px;">
                        <span style="font-size: 14px;">+</span>
                    </button>
                    {self.options["title"]}
                </div>
        """
        if self.options.get("link", None):
            raw += f"""
             <span class="badge gs-badge-link">
             <a class="reference external" href="{self.options["link"]}">Straight to tutorial…</a>
             </span>
             """
        raw += f"""

            </div>
            <div id="{self.options["id"]}" class="collapse" data-parent="#accordion">
                <div class="card-body">
        """
        include_lines = raw.splitlines()
        c = Raw(
            self.name,
            ["html"],
            self.options,
            include_lines,  # content
            self.lineno,
            self.content_offset,
            self.block_text,
            self.state,
            self.state_machine,
        )
        return c.run()


class AccordionFooter(Directive):
    """Insert HTML code to close an accordion box in the How To."""

    def run(self):
        raw = """
                    </div>
                </div>
            </div>
        </div>
        """
        include_lines = raw.splitlines()
        c = Raw(
            self.name,
            ["html"],
            self.options,
            include_lines,  # content
            self.lineno,
            self.content_offset,
            self.block_text,
            self.state,
            self.state_machine,
        )
        return c.run()


class HowtoHLI(Include):
    """Directive to insert how-to for high-level interface."""

    def run(self):
        raw = ""
        section = self.arguments[0]
        doc = AnalysisConfig._get_doc_sections()
        for keyword in doc.keys():
            if section in ["", keyword]:
                raw += doc[keyword]
        include_lines = raw.splitlines()
        codeblock = CodeBlock(
            self.name,
            [],
            self.options,
            include_lines,  # content
            self.lineno,
            self.content_offset,
            self.block_text,
            self.state,
            self.state_machine,
        )
        return codeblock.run()


class DocsImage(Image):
    """Directive to add optional images from gammapy-data."""

    def run(self):
        filename = self.arguments[0]

        if HAS_GP_DATA:
            path = gammapy_data_path / "figures" / filename
            if not path.is_file():
                msg = "Error in {} directive: File not found: {}".format(
                    self.name, path
                )
                raise self.error(msg)
            # Usually Sphinx doesn't support absolute paths
            # But passing a POSIX string of the absolute path
            # with an extra "/" at the start seems to do the trick
            self.arguments[0] = "/" + path.absolute().as_posix()
        else:
            self.warning(
                "GAMMAPY_DATA not available. "
                "Missing image: name: {!r} filename: {!r}".format(self.name, filename)
            )
            self.options["alt"] = self.arguments[1] if len(self.arguments) > 1 else ""

        return super().run()


class SubstitutionCodeBlock(CodeBlock):
    """Similar to CodeBlock but replaces placeholders with variables."""

    def run(self):
        """Replace placeholders with given variables."""
        app = self.state.document.settings.env.app
        new_content = []
        self.content = self.content
        existing_content = self.content
        for item in existing_content:
            for pair in app.config.substitutions:
                original, replacement = pair
                item = item.replace(original, replacement)
            new_content.append(item)

        self.content = new_content
        return list(CodeBlock.run(self))


class DynamicPRLinkTransform(SphinxPostTransform):
    """A Sphinx post-transform that converts [#XXXX] into clickable links to GitHub PR pages."""

    default_priority = 800
    PR_PATTERN = re.compile(r"\[#(\d+)\]")  # Matches [#XXXX]
    GITHUB_PR_BASE_URL = "https://github.com/gammapy/gammapy/pull/"  # Fixed URL

    def apply(self):
        """Apply the transform to convert PR references into links."""
        for text_node in self.document.traverse(nodes.Text):
            if "[#" not in text_node:  # Skip nodes unlikely to match
                continue

            content, last_idx = [], 0
            # Find the associated pattern
            for match in self.PR_PATTERN.finditer(text_node.astext()):
                # Add preceding text
                content.append(nodes.Text(text_node[last_idx : match.start()]))

                # Create hyperlink node for the PR
                pr_number = match.group(1)
                pr_url = f"{self.GITHUB_PR_BASE_URL}{pr_number}"
                link_node = nodes.reference(text=f"[#{pr_number}]", refuri=pr_url)
                content.append(link_node)

                last_idx = match.end()

            # Add any remaining text after the last match
            content.append(nodes.Text(text_node[last_idx:]))

            # Replace the original text node with the new processed nodes
            text_node.parent.replace(text_node, content)


def gammapy_sphinx_ext_activate():
    if HAS_GP_DATA:
        log.info(f"*** Found GAMMAPY_DATA = {gammapy_data_path}")
        log.info("*** Nice!")
    else:
        log.info("*** gammapy-data *not* found.")
        log.info("*** Set the GAMMAPY_DATA environment variable!")
        log.info("*** Docs build will be incomplete.")

    # Register our directives and roles with Sphinx
    register_directive("gp-image", DocsImage)
    register_directive("gp-howto-hli", HowtoHLI)
    register_directive("accordion-header", AccordionHeader)
    register_directive("accordion-footer", AccordionFooter)



"""
CTAO with Gammapy
=================

Access and inspect CTAO data and instrument response functions (IRFs) using Gammapy.

Introduction
------------

The `Cherenkov Telescope Array Observatory (CTAO) <https://www.ctao.org/>`__ is the next generation
ground-based observatory for gamma-ray astronomy. Gammapy is the core
library for the Cherenkov Telescope Array Observatory (CTAO) science tools
(`2017ICRC…35..766D <https://ui.adsabs.harvard.edu/abs/2017ICRC...35..766D>`__
and `CTAO Press
Release <https://www.ctao.org/news/ctao-adopts-the-gammapy-software-package-for-science-analysis/>`__).

CTAO will start taking data in the coming years. For now, to learn how to
analyse CTAO data and to use Gammapy, if you are a member of the CTAO
consortium, you can use the simulated dataset from:

- the CTA first data challenge which ran in 2017 and 2018 (see `here <https://forge.in2p3.fr/projects/data-challenge-1-dc-1/wiki>`__
  for CTAO members)
- the CTAO Science Data Challenge of 2024 (see `here <https://ctaoobservatory.sharepoint.com/:f:/r/sites/share-open-data/Shared%20Documents/Reference%20datasets/Internal%20Science%20Data%20Challenge?csf=1&web=1&e=gNuFzI>`__
  for CTAO members)

Gammapy fully supports the FITS data formats (events, IRFs) used in CTA
1DC and SDC. The XML sky model format is not supported, but are also not needed
to analyse the data, you have to specify your model via the Gammapy YAML
model format, or Python code, as shown below.

You can also use Gammapy to simulate CTAO data and evaluate CTAO performance
using the CTAO response files. Two sets of responses are available for different
array layouts:

- the Omega configuration (prod3b, 2016):  https://zenodo.org/records/5163273,
- the Alpha configuration (prod5, 2021): https://zenodo.org/records/5499840.

They are all fully supported by Gammapy.


Tutorial overview
-----------------

This notebook shows how to access CTAO data and instrument response
functions (IRFs) using Gammapy, and gives some examples how to quick
look the content of CTAO files, especially to see the shape of CTAO IRFs.

At the end of the notebooks, we give several links to other tutorial
notebooks that show how to simulate CTAO data and how to evaluate CTAO
observability and sensitivity, or how to analyse CTAO data.

Note that the FITS data and IRF format currently used by CTAO is the one
documented at https://gamma-astro-data-formats.readthedocs.io/, and is
also used by H.E.S.S. and other Imaging Atmospheric Cherenkov Telescopes
(IACTs). So if you see other Gammapy tutorials using e.g. H.E.S.S.
example data, know that they also apply to CTAO, all you have to do is to
change the loaded data or IRFs to CTAO.

Setup
-----

"""

import os
from pathlib import Path
from astropy import units as u

# %matplotlib inline
import matplotlib.pyplot as plt
from IPython.display import display
from gammapy.data import DataStore, EventList
from gammapy.irf import EffectiveAreaTable2D, load_irf_dict_from_file


######################################################################
# CTA 1DC
# -------
#
# The CTA first data challenge (1DC) ran in 2017 and 2018. It is described
# in detail
# `here <https://forge.in2p3.fr/projects/data-challenge-1-dc-1/wiki>`__
# and a description of the data and how to download it is
# `here <https://forge.in2p3.fr/projects/data-challenge-1-dc-1/wiki#Data-access>`__.
#
# You should download `caldb.tar.gz` (1.2 MB), `models.tar.gz` (0.9
# GB), `index.tar.gz` (0.5 MB), as well as optionally the simulated
# survey data you are interested in: Galactic plane survey `gps.tar.gz`
# (8.3 GB), Galactic center `gc.tar.gz` (4.4 MB), Extragalactic survey
# `egal.tar.gz` (2.5 GB), AGN monitoring `agn.wobble.tar.gz` (4.7 GB).
# After download, follow the instructions how to `untar` the files, and
# set a `CTADATA` environment variable to point to the data.
#
# **For convenience**, since the 1DC data files are large and not publicly
# available to anyone, we have taken a tiny subset of the CTA 1DC data,
# four observations with the southern array from the GPS survey, pointing
# near the Galactic center, and **included them at** `$GAMMAPY_DATA/cta-1dc`.
#
# Files
# ~~~~~
#
# Next we will show a quick overview of the files and how to load them,
# and some quick look plots showing the shape of the CTAO IRFs. How to do
# CTAO simulations and analyses is shown in other tutorials, see links at
# the end of this notebook.
#

# !ls -1 $GAMMAPY_DATA/cta-1dc

# !ls -1 $GAMMAPY_DATA/cta-1dc/data/baseline/gps

# !ls -1 $GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h

# !ls -1 $GAMMAPY_DATA/cta-1dc/index/gps


######################################################################
# The access to the IRFs files requires to define a `CALDB` environment
# variable. We are going to define it only for this notebook so it won’t
# overwrite the one you may have already defined.
#

os.environ["CALDB"] = os.environ["GAMMAPY_DATA"] + "/cta-1dc/caldb"


######################################################################
# Datastore
# ~~~~~~~~~
#
# You can use the `~gammapy.data.DataStore` to load via the index files
#

data_store = DataStore.from_dir("$GAMMAPY_DATA/cta-1dc/index/gps")
print(data_store)


######################################################################
# If you can’t download the index files, or got errors related to the data
# access using them, you can generate the `~gammapy.data.DataStore` directly from the
# event files.
#

path = Path(os.environ["GAMMAPY_DATA"]) / "cta-1dc/data"
paths = list(path.rglob("*.fits"))
data_store = DataStore.from_events_files(paths)
print(data_store)

data_store.obs_table[["OBS_ID", "GLON_PNT", "GLAT_PNT", "IRF"]]

observation = data_store.obs(110380)
print(observation)


######################################################################
# Events
# ------
#
# We can load events data via the data store and observation, or
# equivalently via the `~gammapy.data.EventList` class by specifying the
# EVENTS filename.
#
# The quick-look `events.peek()` plot below shows that CTAO has a field
# of view of a few degrees, and two energy thresholds, one significantly
# below 100 GeV where the CTAO large-size telescopes (LSTs) detect events,
# and a second one near 100 GeV where the mid-sized telescopes (MSTs)
# start to detect events.
#
# Note that most events are “hadronic background” due to cosmic ray
# showers in the atmosphere that pass the gamma-hadron selection cuts for
# this analysis configuration. Since this is simulated data, column
# `MC_ID` is available that gives an emission component identifier code,
# and the EVENTS header in `events.table.meta` can be used to look up
# which `MC_ID` corresponds to which emission component.
#
# Events can be accessed from the observation object like:

events = observation.events

######################################################################
# Or read directly from an event file:
#

events = EventList.read(
    "$GAMMAPY_DATA/cta-1dc/data/baseline/gps/gps_baseline_110380.fits"
)

######################################################################
# Here we print the data from the first 5 events listed in the table:
#

display(events.table[:5])

######################################################################
# And show a summary plot:
#

events.peek()
plt.show()

######################################################################
# IRFs
# ----
#
# The CTAO instrument response functions (IRFs) are given as FITS files in
# the `caldb` folder, the following IRFs are available:
#
# -  effective area
# -  energy dispersion
# -  point spread function
# -  background
#
# Notes:
#
# -  The IRFs contain the energy and offset dependence of the CTAO response
# -  CTA 1DC was based on an early version of the CTAO FITS responses
#    produced in 2017, improvements have been made since.
# -  The point spread function was approximated by a Gaussian shape
# -  The background is from hadronic and electron air shower events that
#    pass CTAO selection cuts. It was given as a function of field of view
#    coordinates, although it is radially symmetric.
# -  The energy dispersion in CTA 1DC is noisy at low energy, leading to
#    unreliable spectral points for some analyses.
# -  The CTA 1DC response files have the first node at field of view
#    offset 0.5 deg, so to get the on-axis response at offset 0 deg,
#    Gammapy has to extrapolate. Furthermore, because diffuse gamma-rays
#    in the FOV were used to derive the IRFs, and the solid angle at small
#    FOV offset circles is small, the IRFs at the center of the FOV are
#    somewhat noisy. This leads to unstable analysis and simulation issues
#    when using the DC1 IRFs for some analyses.
#

print(observation.aeff)

irf_filename = (
    "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
)
irfs = load_irf_dict_from_file(irf_filename)
print(irfs)


######################################################################
# Effective area
# ~~~~~~~~~~~~~~
#

# Equivalent alternative way to load IRFs directly
aeff = EffectiveAreaTable2D.read(irf_filename, hdu="EFFECTIVE AREA")
print(aeff)

irfs["aeff"].peek()
plt.show()

# What is the on-axis effective area at 10 TeV?
print(aeff.evaluate(energy_true="10 TeV", offset="0 deg").to("km2"))


######################################################################
# Energy dispersion
# ~~~~~~~~~~~~~~~~~
#

irfs["edisp"].peek()
plt.show()


######################################################################
# Point spread function
# ~~~~~~~~~~~~~~~~~~~~~
#

irfs["psf"].peek()
plt.show()


# %%
# This is how for analysis you could slice out the PSF
# at a given field of view offset
plt.figure(figsize=(8, 5))
irfs["psf"].plot_containment_radius_vs_energy(
    offset=[1] * u.deg, fraction=[0.68, 0.8, 0.95]
)
plt.show()

######################################################################
# Background
# ~~~~~~~~~~
#
# The background is given as a rate in units `MeV-1 s-1 sr-1`.
#

irfs["bkg"].peek()
plt.show()

print(irfs["bkg"].evaluate(energy="3 TeV", fov_lon="1 deg", fov_lat="0 deg"))


######################################################################
# To visualise the background at particular energies:
#

irfs["bkg"].plot_at_energy(
    ["100 GeV", "500 GeV", "1 TeV", "3 TeV", "10 TeV", "100 TeV"]
)
plt.show()

######################################################################
# Source models
# -------------
#
# The 1DC sky model is distributed as a set of XML files, which link
# to numerous other FITS and text files. Gammapy does not natively support this
# XML format. Instead, we use YAML-based model format that improves on the XML format
# by being easier to write and read, add
# relevant information (units for physical quantities), and omit useless
# information (e.g. parameter scales in addition to values).
#
# If you need or prefer to work with the original XML model files, you can use tools like
# `ElementTree <https://docs.python.org/3/library/xml.etree.elementtree.html>`__
# from the Python standard library, or
# `xmltodict <https://github.com/martinblech/xmltodict>`__ installed through
# `pip install xmltodict`. Here’s an example how to load the information
# for a given source, and convert it into the sky model format Gammapy
# understands.
#
# With the following command we can see what the XML file looks like
# (uncomment it if you have installed `xmltodict`).
 
 
# !tail -n 20 $CTADATA/models/models_gps.xml

######################################################################
# Here is an example on how to create a gammapy source model from this

# import xmltodict
# from astropy import units as u
# from gammapy.modeling.models import PowerLawSpectralModel, PointSpatialModel, SkyModel

# # Read XML file
# filename = "$CTADATA/models/models_gps.xml"

# # Access spectrum parameters
# data = xmltodict.parse(open(filename).read())
# source0 = data["source_library"]["source"][0]
# spectrum0 = source0["spectrum"]["parameter"]
# spatial0 = source0["spatialModel"]

# # Printing the type shows we have a point-like model
# print(spatial0['@type'])

# # Helper function to get parameter by name
# def get_param(params, name):
#     for p in params:
#         if p["@name"] == name:
#             return float(p["@value"]) * float(p["@scale"])
#     raise KeyError(f"Parameter '{name}' not found.")

# # Helper to convert parameter value * scale
# par_to_val = lambda par: float(par["@value"]) * float(par["@scale"])

# # Create a spectral model
# spectral_model = PowerLawSpectralModel(
#     amplitude=par_to_val(spectrum0[0]) * u.Unit("cm-2 s-1 MeV-1"),
#     index=par_to_val(spectrum0[1]),
#     reference=par_to_val(spectrum0[2]) * u.Unit("MeV"),
# )
# print(spectral_model)

# # Create the spatial model
# spatial_params = spatial0["parameter"]
# spatial_model = PointSpatialModel(
#     lon_0=get_param(spatial_params, "RA") * u.deg,
#     lat_0=get_param(spatial_params, "DEC") * u.deg,
#     frame="icrs"
# )
# print(spatial_model)

# # Create the SkyModel
# sky_model = SkyModel(spectral_model=spectral_model, spatial_model=spatial_model, name=source0['@name'])
# print(sky_model)


######################################################################
# Latest CTAO performance files
# -----------------------------
#
# CTA 1DC is useful to learn how to analyse CTAO data. But to do
# simulations and studies for CTAO now, you should get the most recent CTAO
# IRFs in FITS format from https://www.ctao.org/for-scientists/performance/.
#
# If you want to use other response files, the following code cells (remove the # to uncomment)
# explain how to proceed. This example is made with the Alpha configuration (Prod5).
#

# !curl -o cta-prod5-zenodo-fitsonly-v0.1.zip https://zenodo.org/records/5499840/files/cta-prod5-zenodo-fitsonly-v0.1.zip
# !unzip cta-prod5-zenodo-fitsonly-v0.1.zip
# !ls fits/

# !tar xf fits/CTA-Performance-prod5-v0.1-South-40deg.FITS.tar.gz -C fits/.
# !ls fits/*.fits.gz

# irfs1 = load_irf_dict_from_file("fits/Prod5-South-40deg-SouthAz-14MSTs37SSTs.180000s-v0.1.fits.gz")
# irfs1["aeff"].plot_energy_dependence()

# irfs2 = load_irf_dict_from_file("fits/Prod5-South-40deg-SouthAz-14MSTs37SSTs.1800s-v0.1.fits.gz")
# irfs2["aeff"].plot_energy_dependence()


######################################################################
# Exercises
# ---------
#
# -  Load the EVENTS file for `obs_id=111159` as a
#    `~gammapy.data.EventList` object.
# -  Use ``EventList.table`` to find the energy, sky coordinate and time of
#    the highest-energy event.
# -  Use `~gammapy.data.EventList.pointing_radec` to find the pointing position of this
#    observation, and use `astropy.coordinates.SkyCoord` methods to find
#    the field of view offset of the highest-energy event.
# -  What is the effective area and PSF 68% containment radius of CTAO at 1
#    TeV for the `South_z20_50h` configuration used for the CTA 1DC
#    simulation?
# -  Get the latest CTAO FITS performance files from
#    https://www.ctao.org/for-scientists/performance/ and run the
#    code example above. Make an effective area ratio plot of 40 deg
#    zenith versus 20 deg zenith for the `South_z40_50h` and
#    `South_z20_50h` configurations.
#


######################################################################
# Next steps
# ----------
#
# -  Learn how to analyse data with
#    :doc:`/tutorials/starting/analysis_1` and
#    :doc:`/tutorials/starting/analysis_2` or any other
#    Gammapy analysis tutorial.
# -  Learn how to evaluate CTAO observability and sensitivity with
#    :doc:`/tutorials/analysis-3d/simulate_3d`,
#    :doc:`/tutorials/analysis-1d/spectrum_simulation`
#    or :doc:`/tutorials/analysis-1d/cta_sensitivity`.
#



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import logging
import numpy as np
from astropy import units as u
from astropy.io import fits
from astropy.table import Table
from astropy.visualization import quantity_support
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from gammapy.maps.axes import UNIT_STRING_FORMAT, MapAxis
from gammapy.modeling.models import (
    DatasetModels,
    Models,
    SkyModel,
    TemplateSpatialModel,
)
from gammapy.stats import FIT_STATISTICS_REGISTRY
from gammapy.utils.interpolation import interpolate_profile
from gammapy.utils.scripts import make_name, make_path
from .core import Dataset

log = logging.getLogger(__name__)

__all__ = ["FluxPointsDataset"]


def _get_reference_model(model, energy_bounds, margin_percent=70):
    if isinstance(model.spatial_model, TemplateSpatialModel):
        geom = model.spatial_model.map.geom
        emin = energy_bounds[0] * (1 - margin_percent / 100)
        emax = energy_bounds[-1] * (1 + margin_percent / 100)
        energy_axis = MapAxis.from_energy_bounds(
            emin, emax, nbin=20, per_decade=True, name="energy_true"
        )
        geom = geom.to_image().to_cube([energy_axis])
        return Models([model]).to_template_spectral_model(geom)
    else:
        return model.spectral_model


class FluxPointsDataset(Dataset):
    """Bundle a set of flux points with a parametric model, to compute fit statistic function using different statistics (see ``stat_type``).

    For more information see :ref:`datasets`.

    Parameters
    ----------
    models : `~gammapy.modeling.models.Models`
        Models (only spectral part needs to be set).
    data : `~gammapy.estimators.FluxPoints`
        Flux points. Must be sorted along the energy axis.
    mask_fit : `numpy.ndarray`
        Mask to apply for fitting.
    mask_safe : `numpy.ndarray`
        Mask defining the safe data range. By default, upper limit values are excluded.
    meta_table : `~astropy.table.Table`
        Table listing information on observations used to create the dataset.
        One line per observation for stacked datasets.
    stat_type : str
        Method used to compute the statistics:

                * chi2 : estimate from chi2 statistics.
                * profile : estimate from interpolation of the likelihood profile.
                * distrib : Assuming gaussian errors the likelihood is given by the probability density function
                  of the normal distribution. For the upper limit case it is necessary to marginalize over the unknown
                  measurement, so we integrate the normal distribution up to the upper limit value which gives the
                  complementary error function. See eq. C7 of `Mohanty et al (2013) <https://iopscience.iop.org/article/10.1088/0004-637X/773/2/168/pdf>`__

        Default is `chi2`, in that case upper limits are ignored and the mean of asymetrics error is used.
        However, it is recommended to use `profile` if `stat_scan` is available on flux points.
        The `distrib` case provides an approximation if the profile is not available.
    stat_kwargs : dict
        Extra arguments specifying the interpolation scheme of the likelihood profile.
        Used only if `stat_type=="profile"`. In that case the default is :
        `stat_kwargs={"interp_scale":"sqrt", "extrapolate":True}`

    Examples
    --------
    Load flux points from file and fit with a power-law model::

    >>> from gammapy.modeling import Fit
    >>> from gammapy.modeling.models import PowerLawSpectralModel, SkyModel
    >>> from gammapy.estimators import FluxPoints
    >>> from gammapy.datasets import FluxPointsDataset
    >>> filename = "$GAMMAPY_DATA/tests/spectrum/flux_points/diff_flux_points.fits"
    >>> dataset = FluxPointsDataset.read(filename)
    >>> model = SkyModel(spectral_model=PowerLawSpectralModel())
    >>> dataset.models = model

    Make the fit:

    >>> fit = Fit()
    >>> result = fit.run([dataset])
    >>> print(result)
    OptimizeResult
    <BLANKLINE>
        backend    : minuit
        method     : migrad
        success    : True
        message    : Optimization terminated successfully.
        nfev       : 135
        total stat : 25.21
    <BLANKLINE>
    CovarianceResult
    <BLANKLINE>
        backend    : minuit
        method     : hesse
        success    : True
        message    : Hesse terminated successfully.

    >>> print(result.parameters.to_table())
    type    name     value         unit      ... frozen link prior
    ---- --------- ---------- -------------- ... ------ ---- -----
             index 2.2159e+00                ...  False
         amplitude 2.1619e-13 TeV-1 s-1 cm-2 ...  False
         reference 1.0000e+00            TeV ...   True

    Note: In order to reproduce the example, you need the tests datasets folder.
    You may download it with the command:
    ``gammapy download datasets --tests --out $GAMMAPY_DATA``
    """

    tag = "FluxPointsDataset"

    def __init__(
        self,
        models=None,
        data=None,
        mask_fit=None,
        mask_safe=None,
        name=None,
        meta_table=None,
        stat_type="chi2",
        stat_kwargs=None,
    ):
        if not data.geom.has_energy_axis:
            raise ValueError("FluxPointsDataset needs an energy axis")
        self.data = data
        self.mask_fit = mask_fit
        self._name = make_name(name)
        self.models = models
        self.meta_table = meta_table

        if stat_kwargs is None:
            stat_kwargs = dict()
        self.stat_kwargs = stat_kwargs

        self.stat_type = stat_type

        if mask_safe is None:
            mask_safe = np.ones(self.data.dnde.data.shape, dtype=bool)
        self.mask_safe = mask_safe

    @property
    def available_stat_type(self):
        return ["chi2", "distrib", "profile"]

    @property
    def stat_type(self):
        return self._stat_type

    @stat_type.setter
    def stat_type(self, stat_type):
        if stat_type not in self.available_stat_type:
            raise ValueError(
                f"Invalid stat_type: possible options are {self.available_stat_type}"
            )

        if stat_type == "chi2":
            self._mask_valid = (~self.data.is_ul).data & np.isfinite(self.data.dnde)
        elif stat_type == "distrib":
            self._mask_valid = (
                self.data.is_ul.data & np.isfinite(self.data.dnde_ul)
            ) | np.isfinite(self.data.dnde)
        elif stat_type == "profile":
            self.stat_kwargs.setdefault("interp_scale", "sqrt")
            self.stat_kwargs.setdefault("extrapolate", True)
            self._profile_interpolators = self._get_valid_profile_interpolators()
        self._stat_type = stat_type
        self._fit_statistic = FIT_STATISTICS_REGISTRY[stat_type]

    @property
    def mask_valid(self):
        return self._mask_valid

    @property
    def mask_safe(self):
        return self._mask_safe & self.mask_valid

    @mask_safe.setter
    def mask_safe(self, mask_safe):
        self._mask_safe = mask_safe

    @property
    def name(self):
        return self._name

    @property
    def gti(self):
        """Good time interval info (`GTI`)."""
        return self.data.gti

    @property
    def models(self):
        return self._models

    @models.setter
    def models(self, models):
        if models is None:
            self._models = None
        else:
            models = DatasetModels(models)
            self._models = models.select(datasets_names=self.name)

    def write(self, filename, overwrite=False, checksum=False, **kwargs):
        """Write flux point dataset to file.

        Parameters
        ----------
        filename : str
            Filename to write to.
        overwrite : bool, optional
            Overwrite existing file. Default is False.
        checksum : bool
            When True adds both DATASUM and CHECKSUM cards to the headers written to the FITS file.
            Applies only if filename has .fits suffix. Default is False.
        **kwargs : dict, optional
             Keyword arguments passed to `~astropy.table.Table.write`.
        """
        table = self.data.to_table()

        if self.mask_fit is None:
            mask_fit = self.mask_safe
        else:
            mask_fit = self.mask_fit

        table["mask_fit"] = mask_fit
        table["mask_safe"] = self.mask_safe

        filename = make_path(filename)

        if "fits" in filename.suffixes:
            primary_hdu = fits.PrimaryHDU()
            hdu_table = fits.BinTableHDU(table, name="TABLE")
            fits.HDUList([primary_hdu, hdu_table]).writeto(
                filename, overwrite=overwrite, checksum=checksum
            )
        else:
            table.write(make_path(filename), overwrite=overwrite, **kwargs)

    @classmethod
    def read(cls, filename, name=None):
        """Read pre-computed flux points and create a dataset.

        Parameters
        ----------
        filename : str
            Filename to read from.
        name : str, optional
            Name of the new dataset. Default is None.

        Returns
        -------
        dataset : `FluxPointsDataset`
            FluxPointsDataset.
        """
        from gammapy.estimators import FluxPoints

        filename = make_path(filename)
        table = Table.read(filename)
        mask_fit = None
        mask_safe = None

        if "mask_safe" in table.colnames:
            mask_safe = table["mask_safe"].data.astype("bool")

        if "mask_fit" in table.colnames:
            mask_fit = table["mask_fit"].data.astype("bool")

        return cls(
            name=make_name(name),
            data=FluxPoints.from_table(table),
            mask_fit=mask_fit,
            mask_safe=mask_safe,
        )

    @classmethod
    def from_dict(cls, data, **kwargs):
        """Create flux point dataset from dictionary.

        Parameters
        ----------
        data : dict
            Dictionary containing data to create dataset from.

        Returns
        -------
        dataset : `FluxPointsDataset`
            Flux point datasets.
        """
        from gammapy.estimators import FluxPoints

        filename = make_path(data["filename"])
        table = Table.read(filename)
        mask_fit = table["mask_fit"].data.astype("bool")
        mask_safe = table["mask_safe"].data.astype("bool")
        table.remove_columns(["mask_fit", "mask_safe"])
        return cls(
            name=data["name"],
            data=FluxPoints.from_table(table),
            mask_fit=mask_fit,
            mask_safe=mask_safe,
        )

    def __str__(self):
        str_ = f"{self.__class__.__name__}\n"
        str_ += "-" * len(self.__class__.__name__) + "\n"
        str_ += "\n"

        str_ += "\t{:32}: {} \n\n".format("Name", self.name)

        # data section
        n_bins = 0
        if self.data is not None:
            n_bins = np.prod(self.data.geom.data_shape)
        str_ += "\t{:32}: {} \n".format("Number of total flux points", n_bins)

        n_fit_bins = 0
        if self.mask is not None:
            n_fit_bins = np.sum(self.mask.data)
        str_ += "\t{:32}: {} \n\n".format("Number of fit bins", n_fit_bins)

        # likelihood section
        str_ += "\t{:32}: {}\n".format("Fit statistic type", self.stat_type)

        stat = np.nan
        if self.data is not None and self.models is not None:
            stat = self.stat_sum()
        str_ += "\t{:32}: {:.2f}\n\n".format("Fit statistic value (-2 log(L))", stat)

        # model section
        n_models = 0
        if self.models is not None:
            n_models = len(self.models)

        str_ += "\t{:32}: {} \n".format("Number of models", n_models)

        if self.models is not None:
            str_ += "\t{:32}: {}\n".format(
                "Number of parameters", len(self.models.parameters)
            )
            str_ += "\t{:32}: {}\n\n".format(
                "Number of free parameters", len(self.models.parameters.free_parameters)
            )
            str_ += "\t" + "\n\t".join(str(self.models).split("\n")[2:])

        return str_.expandtabs(tabsize=2)

    def data_shape(self):
        """Shape of the flux points data (tuple)."""
        return self.data.energy_ref.shape

    def flux_pred(self):
        """Compute predicted flux."""
        flux = 0.0
        for model in self.models:
            reference_model = _get_reference_model(model, self._energy_bounds)
            sky_model = SkyModel(
                spectral_model=reference_model, temporal_model=model.temporal_model
            )
            flux_model = sky_model.evaluate_geom(
                self.data.geom.as_energy_true, self.gti
            )
            flux += flux_model
        return flux

    def _get_valid_profile_interpolators(self):
        value_scan = self.data.stat_scan.geom.axes["norm"].center
        shape_axes = self.data.stat_scan.geom._shape[slice(3, None)][::-1]
        interpolators = np.empty(shape_axes, dtype=object)
        self._mask_valid = np.ones(self.data.dnde.data.shape, dtype=bool)
        for idx in np.ndindex(shape_axes):
            stat_scan = np.abs(
                self.data.stat_scan.data[idx].squeeze()
                - self.data.stat.data[idx].squeeze()
            )
            self._mask_valid[idx] = np.all(np.isfinite(stat_scan))
            interpolators[idx] = interpolate_profile(
                value_scan,
                stat_scan,
                interp_scale=self.stat_kwargs["interp_scale"],
                extrapolate=self.stat_kwargs["extrapolate"],
            )
        return interpolators

    def residuals(self, method="diff"):
        """Compute flux point residuals.

        Parameters
        ----------
        method: {"diff", "diff/model"}
            Method used to compute the residuals. Available options are:

            - ``"diff"`` (default): data - model.
            - ``"diff/model"``: (data - model) / model.

        Returns
        -------
        residuals : `~numpy.ndarray`
            Residuals array.
        """
        fp = self.data

        model = self.flux_pred()

        residuals = self._compute_residuals(fp.dnde.quantity, model, method)
        # Remove residuals for upper_limits
        residuals[fp.is_ul.data] = np.nan
        return residuals

    def plot_fit(
        self,
        ax_spectrum=None,
        ax_residuals=None,
        kwargs_spectrum=None,
        kwargs_residuals=None,
    ):
        """Plot flux points, best fit model and residuals in two panels.

        Calls `~FluxPointsDataset.plot_spectrum` and `~FluxPointsDataset.plot_residuals`.

        Parameters
        ----------
        ax_spectrum : `~matplotlib.axes.Axes`, optional
            Axes to plot flux points and best fit model on. Default is None.
        ax_residuals : `~matplotlib.axes.Axes`, optional
            Axes to plot residuals on. Default is None.
        kwargs_spectrum : dict, optional
            Keyword arguments passed to `~FluxPointsDataset.plot_spectrum`. Default is None.
        kwargs_residuals : dict, optional
            Keyword arguments passed to `~FluxPointsDataset.plot_residuals`. Default is None.

        Returns
        -------
        ax_spectrum, ax_residuals : `~matplotlib.axes.Axes`
            Flux points, best fit model and residuals plots.

        Examples
        --------
        >>> from gammapy.modeling.models import PowerLawSpectralModel, SkyModel
        >>> from gammapy.estimators import FluxPoints
        >>> from gammapy.datasets import FluxPointsDataset

        >>> #load precomputed flux points
        >>> filename = "$GAMMAPY_DATA/tests/spectrum/flux_points/diff_flux_points.fits"
        >>> flux_points = FluxPoints.read(filename)
        >>> model = SkyModel(spectral_model=PowerLawSpectralModel())
        >>> dataset = FluxPointsDataset(model, flux_points)
        >>> #configuring optional parameters
        >>> kwargs_spectrum = {"kwargs_model": {"color":"red", "ls":"--"}, "kwargs_fp":{"color":"green", "marker":"o"}}
        >>> kwargs_residuals = {"color": "blue", "markersize":4, "marker":'s', }
        >>> dataset.plot_fit(kwargs_residuals=kwargs_residuals, kwargs_spectrum=kwargs_spectrum) # doctest: +SKIP
        """
        if self.data.geom.ndim > 3:
            raise ValueError("Plot fit works with only one energy axis")
        fig = plt.figure(figsize=(9, 7))

        gs = GridSpec(7, 1)
        if ax_spectrum is None:
            ax_spectrum = fig.add_subplot(gs[:5, :])
            if ax_residuals is None:
                plt.setp(ax_spectrum.get_xticklabels(), visible=False)

        if ax_residuals is None:
            ax_residuals = fig.add_subplot(gs[5:, :], sharex=ax_spectrum)

        kwargs_spectrum = kwargs_spectrum or {}
        kwargs_residuals = kwargs_residuals or {}
        kwargs_residuals.setdefault("method", "diff/model")

        self.plot_spectrum(ax=ax_spectrum, **kwargs_spectrum)
        self.plot_residuals(ax=ax_residuals, **kwargs_residuals)
        return ax_spectrum, ax_residuals

    @property
    def _energy_bounds(self):
        try:
            return u.Quantity([self.data.energy_min.min(), self.data.energy_max.max()])
        except KeyError:
            return u.Quantity([self.data.energy_ref.min(), self.data.energy_ref.max()])

    @property
    def _energy_unit(self):
        return self.data.energy_ref.unit

    def plot_residuals(self, ax=None, method="diff", **kwargs):
        """Plot flux point residuals.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional
            Axes to plot on. Default is None.
        method : {"diff", "diff/model"}
            Normalization used to compute the residuals, see `FluxPointsDataset.residuals`. Default is "diff".
        **kwargs : dict
            Keyword arguments passed to `~matplotlib.axes.Axes.errorbar`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axes object.

        """
        if self.data.geom.ndim > 3:
            raise ValueError("Plot residuals works with only one energy axis")
        ax = ax or plt.gca()

        fp = self.data
        residuals = self.residuals(method)

        xerr = self.data.energy_axis.as_plot_xerr

        yerr = fp._plot_get_flux_err(sed_type="dnde")

        if method == "diff/model":
            model = self.flux_pred()
            yerr = (
                (yerr[0].quantity / model).squeeze(),
                (yerr[1].quantity / model).squeeze(),
            )
        elif method == "diff":
            yerr = yerr[0].quantity.squeeze(), yerr[1].quantity.squeeze()
        else:
            raise ValueError('Invalid method, choose between "diff" and "diff/model"')

        kwargs.setdefault("color", kwargs.pop("c", "black"))
        kwargs.setdefault("marker", "+")
        kwargs.setdefault("linestyle", kwargs.pop("ls", "none"))

        with quantity_support():
            ax.errorbar(
                fp.energy_ref, residuals.squeeze(), xerr=xerr, yerr=yerr, **kwargs
            )

        ax.axhline(0, color=kwargs["color"], lw=0.5)

        # format axes
        ax.set_xlabel(f"Energy [{self._energy_unit.to_string(UNIT_STRING_FORMAT)}]")
        ax.set_xscale("log")
        label = self._residuals_labels[method]
        ax.set_ylabel(f"Residuals\n {label}")
        ymin = np.nanmin(residuals - yerr[0])
        ymax = np.nanmax(residuals + yerr[1])
        ymax = max(abs(ymin), ymax)
        ax.set_ylim(-1.05 * ymax, 1.05 * ymax)
        return ax

    def plot_spectrum(
        self, ax=None, kwargs_fp=None, kwargs_model=None, axis_name="energy"
    ):
        """Plot flux points and model.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional
            Axes to plot on. Default is None.
        kwargs_fp : dict, optional
            Keyword arguments passed to `gammapy.estimators.FluxPoints.plot` to configure the plot style.
            Default is None.
        kwargs_model : dict, optional
            Keyword arguments passed to `gammapy.modeling.models.SpectralModel.plot` and
            `gammapy.modeling.models.SpectralModel.plot_error` to configure the plot style. Default is None.
        axis_name : str
            Axis along which to plot the flux points for multiple axes. Default is energy.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axes object.

        Examples
        --------
        >>> from gammapy.modeling.models import PowerLawSpectralModel, SkyModel
        >>> from gammapy.estimators import FluxPoints
        >>> from gammapy.datasets import FluxPointsDataset

        >>> #load precomputed flux points
        >>> filename = "$GAMMAPY_DATA/tests/spectrum/flux_points/diff_flux_points.fits"
        >>> flux_points = FluxPoints.read(filename)
        >>> model = SkyModel(spectral_model=PowerLawSpectralModel())
        >>> dataset = FluxPointsDataset(model, flux_points)
        >>> #configuring optional parameters
        >>> kwargs_model = {"color":"red", "ls":"--"}
        >>> kwargs_fp = {"color":"green", "marker":"o"}
        >>> dataset.plot_spectrum(kwargs_fp=kwargs_fp, kwargs_model=kwargs_model) # doctest: +SKIP
        """
        kwargs_fp = (kwargs_fp or {}).copy()
        kwargs_model = (kwargs_model or {}).copy()

        # plot flux points
        kwargs_fp.setdefault("sed_type", "e2dnde")
        if axis_name == "time":
            kwargs_fp["sed_type"] = "norm"
        ax = self.data.plot(ax=ax, **kwargs_fp, axis_name=axis_name)

        kwargs_model.setdefault("label", "Best fit model")

        kwargs_model.setdefault("zorder", 10)

        for model in self.models:
            if model.datasets_names is None or self.name in model.datasets_names:
                if axis_name == "energy":
                    kwargs_model.setdefault("sed_type", "e2dnde")
                    kwargs_model.setdefault("energy_bounds", self._energy_bounds)
                    model.spectral_model.plot(ax=ax, **kwargs_model)
                if axis_name == "time":
                    kwargs_model.setdefault(
                        "time_range", self.data.geom.axes["time"].time_bounds
                    )
                    model.temporal_model.plot(ax=ax, **kwargs_model)

        if axis_name == "energy":
            kwargs_model["color"] = ax.lines[-1].get_color()
            kwargs_model.pop("label")
            for model in self.models:
                if model.datasets_names is None or self.name in model.datasets_names:
                    model.spectral_model.plot_error(ax=ax, **kwargs_model)

        return ax



"""
Fermi-LAT with Gammapy
======================

Data inspection and preliminary analysis with Fermi-LAT data.

Introduction
------------

Gammapy fully supports Fermi-LAT data analysis from DL4 level (binned
maps). In order to perform data reduction from the events list and
spacecraft files to binned counts and IRFs maps we recommend to use
`Fermipy <http://fermipy.readthedocs.io/>`__, which is based on
the `Fermi Science
Tools <https://fermi.gsfc.nasa.gov/ssc/data/analysis/software/>`__
(Fermi ST).

Using Gammapy with Fermi-LAT data could be an option for you if you want
to do an analysis that is not easily possible with Fermipy and the Fermi
Science Tools. For example a joint likelihood fit of Fermi-LAT data with
data e.g. from H.E.S.S., MAGIC, VERITAS or some other instrument, or
analysis of Fermi-LAT data with a complex spatial or spectral model that
is not available in Fermipy or the Fermi ST.

This tutorial will show you how to convert Fermi-LAT data into a DL4
format  that can be used by Gammapy (`~gammapy.datasets.MapDataset`) and perform a 3D analysis. As an
example, we will look at the Galactic center.
We are going to analyses high-energy data from 10 GeV from 1 TeV (in reconstructed energy).

Setup
-----

For this notebook you have to get the prepared
`fermi-gc` data provided in your `$GAMMAPY_DATA`.


"""

# %matplotlib inline

import numpy as np
from astropy import units as u
import matplotlib.pyplot as plt

from gammapy.catalog import CATALOG_REGISTRY
from gammapy.datasets import Datasets, FermipyDatasetsReader
from gammapy.estimators import TSMapEstimator
from gammapy.maps import Map
from gammapy.modeling import Fit
from gammapy.modeling.models import (
    Models,
    PointSpatialModel,
    PowerLawSpectralModel,
    TemplateSpatialModel,
    SkyModel,
    PowerLawNormSpectralModel,
)
from gammapy.utils.scripts import make_path

######################################################################
# Check setup
# ~~~~~~~~~~~
#
# We check the setup in this tutorial, as we require specific files to be
# downloaded to continue.

from gammapy.utils.check import check_tutorials_setup

check_tutorials_setup()


######################################################################
# Fermipy configuration file
# --------------------------
#
# Gammapy can utilise the same configuration file as Fermipy to convert
# the Fermipy-generated maps into Gammapy datasets. For more information on the
# structure of these files, refer to the `Fermipy configuration
# page <https://fermipy.readthedocs.io/en/latest/config.html>`__. In this
# tutorial, we will analyse Galactic center data generated with Fermipy version 1.3 and
# the configuration given in
# `$GAMMAPY_DATA/fermi-gc/config_fermipy_gc_example.yaml`:
#


######################################################################
# .. code:: yaml
#
#    # Fermipy example configuration
#    # for details, see https://fermipy.readthedocs.io/en/latest/config.html
#    # For IRFs, event type and event class options, see https://fermi.gsfc.nasa.gov/ssc/data/analysis/documentation/Cicerone/Cicerone_Data/LAT_DP.html
#    components:
#      - model: {isodiff: $FERMI_DIR/refdata/fermi/galdiffuse/iso_P8R3_CLEAN_V3_PSF2_v1.txt}
#        selection: {evtype: 16}  #4 is PSF0, 8 PSF1, 16 PSF2, 32 PSF3
#        data: {ltcube: null}
#      - model: {isodiff: $FERMI_DIR/refdata/fermi/galdiffuse/iso_P8R3_CLEAN_V3_PSF3_v1.txt}
#        selection: {evtype: 32}
#        data: {ltcube: null}
#
#    data:
#      evfile : ./raw/events_list.lst
#      scfile : ./raw/L241227031840F357373F12_SC00.fits
#
#    binning:
#      roiwidth   : 8.0
#      binsz      : 0.1
#      binsperdec   : 10
#      coordsys : GAL
#      proj: CAR
#      projtype: WCS
#
#    selection :
#    # gtselect parameters
#      emin : 3981.0717055349733 # ENERGY TRUE for Gammapy
#      emax : 2511886.4315095823 # ENERGY TRUE for Gammapy
#      zmax    : 105 # deg
#      evclass : 256 # CLEAN
#      tmin    : 239557417
#      tmax    : 752112005
#
#    # gtmktime parameters
#      filter : 'DATA_QUAL>0 && LAT_CONFIG==1'
#      roicut : 'no'
#
#    # Set the ROI center to the coordinates of this source
#      glon : 0.
#      glat : 0.
#
#    fileio:
#       outdir : ''
#       logfile : 'out.log'
#       usescratch : False
#       scratchdir  : '/scratch'
#
#    gtlike:
#      edisp : True
#      edisp_bins : 0 # DO NOT CHANGE edisp_bins will be handled by Gammapy
#      irfs : 'P8R3_CLEAN_V3'
#
#    model:
#      src_roiwidth : 10.0 # This is used by Fermipy to compute the PSF RADMAX, even if no models are set
#


######################################################################
# The most important points for Gammapy users are:
#
# - ``emin`` and ``emax`` in this file should be considered as the energy true range.
#   It should be larger that the reconstructed energy range.
# - ``edisp_bins : 0`` is strongly recommended at this stage otherwise you
#   might face inconsistencies in the energy axes of the different IRFs created by Fermipy.
# - The ``edisp_bins`` value will be redefined later on by Gammapy as a positive value
#   in order to create the reconstructed energy axis properly.
# - If you want to use the `$FERMI_DIR` variable to read the background models
#   it must also be defined in your Gammapy environment,
#   otherwise you have to define your own paths.
# - For this tutorial we copied the iso files in `$GAMMAPY_DATA/fermi-gc` and
#   edited the paths in the yaml file for simplicity.
#
# More generally in order to select a good binning it is important to know
# the instrument resolution, for that you can have a quick look at the
# IRFs in the `Fermi-LAT performance
# page <https://www.slac.stanford.edu/exp/glast/groups/canda/lat_Performance.htm>`__.
#
# Since the energy resolution varies with energy, it is important to
# choose an energy binning that is fine enough to capture this energy
# dependence. That is why we recommend a binning with 8 to 10 bins per
# decade. The energy axes will be created such as it is linear in log space
# so it's better to define ``emin`` and ``emax`` such as they align with a log binning.
# Here we have as true energy range :math:`\log(emin) = 0.6 \sim 4` GeV to
# :math:`\log(emax) = 3.4 \sim 2500` GeV.
# While the reconstructed energy range of our analysis will be 10 GeV to 1000 GeV.
#
# The spatial binning should be of the same order of the PSF 68%
# containment radius which is in average 0.1 degree above 10 GeV and
# rapidly increases at lower energy. Ideally it should remain within a
# factor of 2 or 3 of the PSF radius at most. In order to properly take
# into account for the sources outside the region of interest that
# contribute inside due to the PSF we have to define a wider ``roiwidth``
# than our actual region of interest. Typically, we need a margin equal to the
# 99% containment of the PSF on each side. Above 10 GeV considering only
# PSF2&3 the 99% PSF containment radius is about 1 degree. Thus, if we
# want to study a 3 degree radius around the GC we have to take a ``roiwidth`` of 8
# deg. (If considering lower energies or including PSF0 and PSF1, it should be much
# larger).
#


######################################################################
# From Fermipy maps to Gammapy datasets
# -------------------------------------
#
# In your Fermipy environment you have to run the following commands
#
# .. code:: python
#
#    from fermipy.gtanalysis import GTAnalysis
#    gta = GTAnalysis('config_fermipy_gc_example.yaml',logging={'verbosity' : 3})
#    gta.setup()
#
#    gta.compute_psf(overwrite=True) # this creates the psf kernel
#    gta.compute_drm(edisp_bins=0, overwrite=True) # this creates the energy dispersion matrix
#    # DO NOT CHANGE edisp_bins here, it will be redefined by Gammapy later on
#
# This will produce a number of files including:
#
# - “ccube_00.fits” (counts)
# - “bexpmap_00.fits” (exposure)
# - “psf_00.fits” (psf)
# - “drm_00.fits” (edisp)
#
#


######################################################################
# In your Gammapy environment you can create the datasets using the
# same configuration file.
#


reader = FermipyDatasetsReader(
    "$GAMMAPY_DATA/fermi-gc/config_fermipy_gc_example.yaml", edisp_bins=4
)
datasets = reader.read()
print(datasets)


######################################################################
# Note that the ``edisp_bins`` is set again here as a positive number so
# Gammapy can create its reconstructed energy axis properly. The energy
# dispersion correction implemented Gammapy is closer to the version
# implemented in Fermitools >1.2.0, which take into account the interplay
# between the energy dispersion and PSF.
#
# Across most of the Fermi energy range, the level of migration in log10(E)
# remains within 0.2, increasing up to 0.4 below 100 MeV,
# due to energy dispersion. Therefore, we recommend that the
# product of \|edisp_bins\| and the width of the log10(E) bins be at
# least equal to 0.2. For a binning of 8 to 10 bins per decade, this
# corresponds to \|edisp_bins\| ≥ 2. For further information, see
# `Pass8_edisp_usage <https://fermi.gsfc.nasa.gov/ssc/data/analysis/documentation/Pass8_edisp_usage.html>`__.
# In our case, we have 10 bins per decade and true energy axis starts
# at about 4 GeV, so with ``edisp_bins=4`` the reconstructed energy axis
# starts at 10 GeV:
#

print(datasets[0].exposure.geom.axes["energy_true"])
print(datasets[0].counts.geom.axes["energy"])

######################################################################
# Note that selecting ``edisp_bins=2`` means the reconstructed energy
# of the counts geometry will start at :math:`10^{0.8} \sim 6.3` GeV.
# If we want to start the analysis at 10 GeV in this case, we need to
# update the ``mask_fit`` to exclude the first 2 reconstructed energy bins.
# Considering more ``edisp_bins`` is generally safer but requires more memory
# and increases computation time.
#
# Alternatively, if you created the counts and IRF files from the
# Fermi-LAT science tools without Fermipy you can use the
# ``create_dataset``  method. Note that in this case we cannot guarantee
# that your maps have the correct axes dimensions to be properly converted
# into Gammapy datasets.
#

path = make_path("$GAMMAPY_DATA/fermi-gc")
dataset0 = reader.create_dataset(
    path / "ccube_00.fits",
    path / "bexpmap_00.fits",
    path / "psf_00.fits",
    path / "drm_00.fits",
    isotropic_file=None,
    edisp_bins=0,
    name="fermi_lat_gc_psf2",
)
dataset1 = reader.create_dataset(
    path / "ccube_01.fits",
    path / "bexpmap_01.fits",
    path / "psf_01.fits",
    path / "drm_01.fits",
    isotropic_file=None,
    edisp_bins=0,
    name="fermi_lat_gc_psf3",
)

datasets_fromST = Datasets([dataset0, dataset1])


# The above was an alternative reading method we don't need those after
del dataset0, dataset1, datasets_fromST


######################################################################
# Fermi-LAT IRF properties
# ------------------------
#
# Exposure
# ~~~~~~~~
#


######################################################################
# Exposure is almost constant across the field of view, with less than 5%
# variation at a given energy.
#

datasets[0].exposure.plot_interactive(add_cbar=True)
plt.show()


######################################################################
# PSF
# ~~~
#
# For Fermi-LAT, the PSF only varies little within a given regions of
# the sky, especially at high energies like what we have here.
# So we have only one PSF kernel.
#

datasets[0].psf.plot_containment_radius_vs_energy(fraction=(0.68, 0.95, 0.99))
plt.show()


######################################################################
# Region of interest and mask definition
# --------------------------------------
#
# As mentioned previously, the width of dataset is larger that our actual
# region of interest in order to properly take into account for the
# sources outside that contributes inside due to the PSF. So we define the
# valid RoI for fitting by creating a ``mask_fit``.
#

margin = (
    2.0 * u.deg
)  # >1 deg should be fine for this dataset we take 2 so the notebook is faster
geom = datasets[0].counts.geom
mask_fit = Map.from_geom(geom, data=True, dtype=bool)
mask_fit = mask_fit.binary_erode(width=margin, kernel="disk")

mask_fit.plot_interactive()
plt.show()

######################################################################
# Now we attach it the datasets
#

for d in datasets:
    d.mask_fit = mask_fit


######################################################################
# Models
# ------
#
# Isotropic diffuse background
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# The `~gammapy.datasets.FermipyDatasetsReader` also created one isotropic diffuse model
# for each dataset:
#

models_iso = Models(datasets.models)
print(models_iso)


######################################################################
# Galactic diffuse background
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
#


######################################################################
# The Fermi-LAT collaboration provides a galactic diffuse emission model,
# that can be used as a background model for Fermi-LAT source analysis.
# These files are called usually IEM for interstellar emission model, the
# latest is
# `gll_iem_v07.fits <https://fermi.gsfc.nasa.gov/ssc/data/analysis/software/aux/4fgl/gll_iem_v07.fits>`__.
# For details see the `BackgroundModels
# page <https://fermi.gsfc.nasa.gov/ssc/data/access/lat/BackgroundModels.html>`__.
# If you have Fermipy installed it can also be found in
# `$FERMI_DIR/refdata/fermi/galdiffuse/gll_iem_v07.fits`
#
# Diffuse model maps are very large (100s of MB), so as an example here,
# we just load one that represents a small cutout for the Galactic center
# region.
#
# In this case, the maps are already in differential units, so we do not
# want to normalise it again.
#

template_iem = TemplateSpatialModel.read(
    filename="$GAMMAPY_DATA/fermi-gc/gll_iem_v07_gc.fits.gz", normalize=False
)

model_iem = SkyModel(
    spectral_model=PowerLawNormSpectralModel(),
    spatial_model=template_iem,
    name="diffuse-iem",
)


######################################################################
# Let’s look at the template :
#

template_iem.map.plot_interactive(add_cbar=True)
plt.show()
# sphinx_gallery_thumbnail_number = 4


models_diffuse = models_iso + model_iem


######################################################################
# Sources
# ~~~~~~~
#
# Source models can be loaded from the 4FGL catalog directly available in
# `$GAMMAPY_DATA`. For details see the `Fermi-LAT catalog
# page <https://fermi.gsfc.nasa.gov/ssc/data/access/lat/14yr_catalog/>`__.
#

catalog_4fgl = CATALOG_REGISTRY.get_cls("4fgl")()  # load 4FGL catalog


######################################################################
# We want to select only the sources inside the dataset spatial geometry:
#

in_geom = geom.to_image().contains(catalog_4fgl.positions)
catalog_4fgl_gc = catalog_4fgl[in_geom]

models_4fgl_gc = catalog_4fgl_gc.to_models()

######################################################################
# That's still quite a lot of sources
#
print("Number of source models", len(models_4fgl_gc))


######################################################################
# In order to improve performances we can store all the sources outside
# the ``mask_fit`` region into a single template (the same could be done
# for all the sources we want to keep frozen).
#

sources_ouside_roi = models_4fgl_gc.select_mask(~mask_fit, use_evaluation_region=False)
sources_inside_roi = Models([m for m in models_4fgl_gc if m not in sources_ouside_roi])

geom_true = datasets[0].exposure.geom
sources_outside_roi = sources_ouside_roi.to_template_sky_model(
    geom_true, name="sources_outside"
)

sources_outside_roi.spatial_model.filename = "sources_outside.fits"

sources_outside_roi.spatial_model.map.plot_interactive(add_cbar=True)
plt.show()

######################################################################
# Now we have less models to describe the sources
#
models_sources = sources_inside_roi + sources_outside_roi

print("Number of source models", len(models_sources))

######################################################################
# Fit
# ---
#
# Now, the big finale: let’s do a 3D of the brightest sources and IEM
# models.
#
# First we attach the models to the datasets.
#

models = models_sources + models_diffuse

datasets.models = models

print("Number of models", len(models))


######################################################################
# Let’s find the 3 brightest sources:
#

n_brightest = 3
integrated_flux = u.Quantity(
    [m.spectral_model.integral(10 * u.GeV, 1 * u.TeV) for m in sources_inside_roi]
)
order = np.argsort(integrated_flux)
selected_sources = Models([sources_inside_roi[int(ii)] for ii in order[:n_brightest]])

print(selected_sources.names)

free_models = selected_sources + model_iem


######################################################################
# We keep only their normalisation free for simplicity:
#

models.freeze()  # freeze all parameters

# and unfreeze only the amplitude or norm of the selected models
for p in free_models.parameters:
    if p.name in ["amplitude", "norm"]:
        p.frozen = False
        p.min = 0

print("Number of free parameters", len(models.parameters.free_parameters))

fit = Fit()
result = fit.run(datasets=datasets)

print(result)


######################################################################
# Residual TS map
# ---------------
#
# Now we can look at the residual TS map to check there is no significant
# excess left:
#

spatial_model = PointSpatialModel()
spectral_model = PowerLawSpectralModel(index=2)
model = SkyModel(spatial_model=spatial_model, spectral_model=spectral_model)

ts_estimator = TSMapEstimator(
    model,
    kernel_width="1 deg",  # this set close to the 95-99% containment radius of the PSF
    selection_optional=[],
    sum_over_energy_groups=True,
    energy_edges=[10, 1000] * u.GeV,
)


ts_results = ts_estimator.run(datasets)


image = ts_results["sqrt_ts"]
image = image.cutout(
    image.geom.center_skydir, width=np.max(image.geom.width) - 2 * margin
)

fig = plt.figure(figsize=(7, 5))
ax = image.plot(
    clim=[-8, 8],
    cmap=plt.cm.RdBu_r,
    add_cbar=True,
    kwargs_colorbar={"label": r"$\sqrt{TS}$ [$\sigma$]"},
)
sources_inside_roi.plot_regions(
    ax=ax, edgecolor="g", linestyle="-", kwargs_point=dict(marker=".")
)
plt.show()


######################################################################
# Serialisation
# -------------
#
# To serialise the created dataset, you must proceed through the Datasets
# API
#

datasets.write(
    filename="fermi_lat_gc_datasets.yaml",
    filename_models="fermi_lat_gc_models.yaml",
    overwrite=True,
)
datasets_read = Datasets.read(
    filename="fermi_lat_gc_datasets.yaml", filename_models="fermi_lat_gc_models.yaml"
)


######################################################################
# Exercises
# ---------
#
# - Fit the position and spectrum of the source `SNR
#   G0.9+0.1 <http://gamma-sky.net/#/cat/tev/110>`__.
# - Make maps and fit the position and spectrum of the `Crab
#   nebula <http://gamma-sky.net/#/cat/tev/25>`__.
#


######################################################################
# Summary
# -------
#
# In this tutorial you have seen how to work with Fermi-LAT data with
# Gammapy. You have to use Fermipy or the Fermi ST to perform the data
# reduction then you can use Gammapy for analysis using the same methods
# that are used to analyse IACT data.
#



"""
Survey Map Script
=================

Make a survey counts map using a script.

We create an all-sky map in AIT projection for the
`H.E.S.S. DL3 DR1 <https://www.mpi-hd.mpg.de/hfm/HESS/pages/dl3-dr1/>`__
dataset.
"""

import logging
from gammapy.data import DataStore
from gammapy.maps import Map

log = logging.getLogger(__name__)


def main():
    data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1")
    obs_id = data_store.obs_table["OBS_ID"]
    observations = data_store.get_observations(obs_id)

    m = Map.create()
    for obs in observations:
        log.info(f"Processing obs_id: {obs.obs_id}")
        m.fill_events(obs.events)

    m.write("survey_map.fits.gz", overwrite=True)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    main()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import pytest
import numpy as np
from numpy.testing import assert_allclose
import astropy.units as u
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
from astropy.units import Quantity
from gammapy.data import (
    DataStore,
    EventList,
    Observation,
    ObservationFilter,
    Observations,
)
from gammapy.data.metadata import ObservationMetaData
from gammapy.data.pointing import FixedPointingInfo
from gammapy.data.utils import get_irfs_features
from gammapy.irf import PSF3D, load_irf_dict_from_file
from gammapy.utils.cluster import hierarchical_clustering
from gammapy.utils.coordinates import FoVICRSFrame, FoVAltAzFrame
from gammapy.utils.fits import HDULocation
from gammapy.utils.testing import (
    assert_skycoord_allclose,
    assert_time_allclose,
    mpl_plot_check,
    requires_data,
)


@pytest.fixture(scope="session")
def data_store():
    return DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/")


@requires_data()
def test_observation(data_store):
    """Test Observation class"""
    obs = data_store.obs(23523)

    assert_time_allclose(obs.tstart, Time(53343.92234009259, scale="tt", format="mjd"))
    assert_time_allclose(obs.tstop, Time(53343.94186555556, scale="tt", format="mjd"))

    c = SkyCoord(83.63333129882812, 21.51444435119629, unit="deg")
    assert_skycoord_allclose(obs.get_pointing_icrs(obs.tmid), c)

    c = SkyCoord(22.558341, 41.950807, unit="deg")
    assert_skycoord_allclose(obs.get_pointing_altaz(obs.tmid), c)

    c = SkyCoord(83.63333129882812, 22.01444435119629, unit="deg")
    assert_skycoord_allclose(obs.target_radec, c)


@requires_data()
def test_observation_get_fov_frame(data_store):
    """Test Observation class"""
    obs = data_store.obs(23523)

    fov_icrs_frame = obs.get_fov_frame(obs.tmid, "RADEC")
    fov_altaz_frame = obs.get_fov_frame(obs.tmid, "ALTAZ")

    pnt_icrs = SkyCoord(83.63333129882812, 21.51444435119629, unit="deg", frame="icrs")

    assert isinstance(fov_icrs_frame, FoVICRSFrame)
    assert_allclose(
        fov_icrs_frame.origin.separation(pnt_icrs).to_value("deg"), 0.0, atol=1e-5
    )

    assert isinstance(fov_altaz_frame, FoVAltAzFrame)
    assert_allclose(fov_altaz_frame.origin.alt.deg, 41.950807)
    assert_allclose(fov_altaz_frame.origin.az.deg, 22.558341)


@requires_data()
def test_observation_peek(data_store):
    obs = Observation.read(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_023523.fits.gz"
    )

    with mpl_plot_check():
        obs.peek()

    obs_with_radmax = Observation.read(
        "$GAMMAPY_DATA/magic/rad_max/data/20131004_05029747_DL3_CrabNebula-W0.40+035.fits"
    )

    with mpl_plot_check():
        obs_with_radmax.peek()


@requires_data()
@pytest.mark.parametrize(
    "time_interval, expected_times",
    [
        (
            Time(
                ["2004-12-04T22:10:00", "2004-12-04T22:30:00"],
                format="isot",
                scale="tt",
            ),
            Time(
                ["2004-12-04T22:10:00", "2004-12-04T22:30:00"],
                format="isot",
                scale="tt",
            ),
        ),
        (
            Time([53343.930, 53343.940], format="mjd", scale="tt"),
            Time([53343.930, 53343.940], format="mjd", scale="tt"),
        ),
        (
            Time([10.0, 100000.0], format="mjd", scale="tt"),
            Time([53343.92234009, 53343.94186563], format="mjd", scale="tt"),
        ),
        (Time([10.0, 20.0], format="mjd", scale="tt"), None),
    ],
)
def test_observation_select_time(data_store, time_interval, expected_times):
    obs = data_store.obs(23523)

    new_obs = obs.select_time(time_interval)

    if expected_times:
        expected_times.format = "mjd"
        assert np.all(
            (new_obs.events.time >= expected_times[0])
            & (new_obs.events.time < expected_times[1])
        )
        assert_time_allclose(new_obs.gti.time_start[0], expected_times[0], atol=0.01)
        assert_time_allclose(new_obs.gti.time_stop[-1], expected_times[1], atol=0.01)
    else:
        assert len(new_obs.events.table) == 0
        assert len(new_obs.gti.table) == 0


@requires_data()
@pytest.mark.parametrize(
    "time_interval, expected_times, expected_nr_of_obs",
    [
        (
            Time([53090.130, 53090.140], format="mjd", scale="tt"),
            Time([53090.130, 53090.140], format="mjd", scale="tt"),
            1,
        ),
        (
            Time([53090.130, 53091.110], format="mjd", scale="tt"),
            Time([53090.130, 53091.110], format="mjd", scale="tt"),
            3,
        ),
        (
            Time([10.0, 53111.0230], format="mjd", scale="tt"),
            Time([53090.1234512, 53111.0230], format="mjd", scale="tt"),
            8,
        ),
        (Time([10.0, 20.0], format="mjd", scale="tt"), None, 0),
    ],
)
def test_observations_select_time(
    data_store, time_interval, expected_times, expected_nr_of_obs
):
    obs_ids = data_store.obs_table["OBS_ID"][:8]
    obss = data_store.get_observations(obs_ids)

    new_obss = obss.select_time(time_interval)

    assert len(new_obss) == expected_nr_of_obs

    if expected_nr_of_obs > 0:
        assert new_obss[0].events.time[0] >= expected_times[0]
        assert new_obss[-1].events.time[-1] < expected_times[1]
        assert_time_allclose(
            new_obss[0].gti.time_start[0], expected_times[0], atol=0.01
        )
        assert_time_allclose(
            new_obss[-1].gti.time_stop[-1], expected_times[1], atol=0.01
        )


@requires_data()
def test_observations_mutation(data_store):
    obs_ids = data_store.obs_table["OBS_ID"][:4]
    obss = data_store.get_observations(obs_ids)
    assert obss.ids == ["20136", "20137", "20151", "20275"]

    obs_id = data_store.obs_table["OBS_ID"][4]
    obs = data_store.get_observations([obs_id])[0]

    obss.append(obs)
    assert obss.ids == ["20136", "20137", "20151", "20275", "20282"]
    obss.insert(0, obs)
    assert obss.ids == ["20282", "20136", "20137", "20151", "20275", "20282"]
    obss.pop(0)
    assert obss.ids == ["20136", "20137", "20151", "20275", "20282"]
    obs3 = obss[3]
    obss.pop(obss.ids[3])
    assert obss.ids == ["20136", "20137", "20151", "20282"]
    obss.insert(3, obs3)
    assert obss.ids == ["20136", "20137", "20151", "20275", "20282"]
    obss.extend([obs])
    assert obss.ids == ["20136", "20137", "20151", "20275", "20282", "20282"]
    obss.remove(obs)
    assert obss.ids == ["20136", "20137", "20151", "20275", "20282"]
    obss[0] = obs
    assert obss.ids == ["20282", "20137", "20151", "20275", "20282"]

    with pytest.raises(TypeError):
        obss.insert(5, "bad")

    with pytest.raises(TypeError):
        obss[5] = "bad"


def test_empty_observations():
    observations = Observations()
    assert len(observations) == 0


@requires_data()
def test_observations_str(data_store):
    obs_ids = data_store.obs_table["OBS_ID"][:4]
    obss = data_store.get_observations(obs_ids)
    actual = obss.__str__()

    assert actual.split("\n")[1] == "Number of observations: 4"


@requires_data()
def test_observations_select_time_time_intervals_list(data_store):
    obs_ids = data_store.obs_table["OBS_ID"][:8]
    obss = data_store.get_observations(obs_ids)
    # third time interval is out of the observations time range
    time_intervals = [
        Time([53090.130, 53090.140], format="mjd", scale="tt"),
        Time([53110.011, 53110.019], format="mjd", scale="tt"),
        Time([53112.345, 53112.42], format="mjd", scale="tt"),
    ]
    new_obss = obss.select_time(time_intervals)

    assert len(new_obss) == 2
    assert new_obss[0].events.time[0] >= time_intervals[0][0]
    assert new_obss[0].events.time[-1] < time_intervals[0][1]
    assert new_obss[1].events.time[0] >= time_intervals[1][0]
    assert new_obss[1].events.time[-1] < time_intervals[1][1]
    assert_time_allclose(new_obss[0].gti.time_start[0], time_intervals[0][0])
    assert_time_allclose(new_obss[0].gti.time_stop[-1], time_intervals[0][1])
    assert_time_allclose(new_obss[1].gti.time_start[0], time_intervals[1][0])
    assert_time_allclose(new_obss[1].gti.time_stop[-1], time_intervals[1][1])


@requires_data()
def test_observation_cta_1dc():
    ontime = 5.0 * u.hr
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(0, 0, unit="deg", frame="galactic").icrs,
    )
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )

    t_ref = Time("2020-01-01T00:00:00")
    tstart = 20 * u.hour
    location = EarthLocation(lon="-70d18m58.84s", lat="-24d41m0.34s", height="2000m")

    obs = Observation.create(
        pointing,
        irfs=irfs,
        deadtime_fraction=0.1,
        tstart=tstart,
        tstop=tstart + ontime,
        reference_time=t_ref,
        location=location,
    )

    assert_skycoord_allclose(obs.get_pointing_icrs(obs.tmid), pointing.fixed_icrs)
    assert_allclose(obs.observation_live_time_duration, 0.9 * ontime)
    assert obs.target_radec is None

    assert isinstance(obs.meta, ObservationMetaData)
    assert obs.meta.deadtime_fraction == 0.1
    assert_allclose(obs.meta.location.height.to_value("m"), 2000)
    assert "Gammapy" in obs.meta.creation.creator


@requires_data()
def test_observation_create_radmax():
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(0, 0, unit="deg", frame="galactic").icrs,
    )
    obs = Observation.read("$GAMMAPY_DATA/joint-crab/dl3/magic/run_05029748_DL3.fits")
    livetime = 5.0 * u.hr
    deadtime = 0.5

    irfs = {}
    for _ in obs.available_irfs:
        irfs[_] = getattr(obs, _)

    obs1 = Observation.create(
        pointing,
        irfs=irfs,
        deadtime_fraction=deadtime,
        livetime=livetime,
    )

    assert_skycoord_allclose(obs1.get_pointing_icrs(obs1.tmid), pointing.fixed_icrs)
    assert_allclose(obs1.observation_live_time_duration, 0.5 * livetime)
    assert obs1.rad_max is not None
    assert obs1.psf is None


@requires_data()
def test_observation_read():
    """read event list and irf components from different DL3 files"""
    obs = Observation.read(
        event_file="$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz",
        irf_file="$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020137.fits.gz",
    )

    energy = Quantity(1, "TeV")
    offset = Quantity(0.5, "deg")
    val = obs.aeff.evaluate(energy_true=energy, offset=offset)

    assert obs.obs_id == 20136
    assert len(obs.events.energy) == 11243
    assert obs.available_hdus == ["events", "gti", "aeff", "edisp", "psf", "bkg"]
    assert_allclose(val.value, 278000.54120855, rtol=1e-5)
    assert val.unit == "m2"

    assert isinstance(obs.meta, ObservationMetaData)
    assert "Gammapy" in obs.meta.creation.creator

    assert obs.meta.obs_info.telescope == "HESS"
    assert obs.meta.obs_info.instrument == "H.E.S.S. Phase I"
    assert obs.meta.target.name == "MSH15-52"
    assert obs.meta.optional["N_TELS"] == 4
    with pytest.raises(KeyError):
        obs.meta.optional["BROKPIX"]


@requires_data()
def test_observation_read_single_file():
    """read event list and irf components from the same DL3 files"""
    obs = Observation.read(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_020136.fits.gz"
    )

    energy = Quantity(1, "TeV")
    offset = Quantity(0.5, "deg")
    val = obs.aeff.evaluate(energy_true=energy, offset=offset)

    assert obs.obs_id == 20136
    assert len(obs.events.energy) == 11243
    assert obs.available_hdus == ["events", "gti", "aeff", "edisp", "psf", "bkg"]
    assert_allclose(val.value, 273372.44851054, rtol=1e-5)
    assert val.unit == "m2"


@requires_data()
def test_observation_read_single_file_fixed_rad_max():
    """check that for a point-like observation without the RAD_MAX_2D table
    a RadMax2D object is generated from the RAD_MAX keyword"""
    obs = Observation.read("$GAMMAPY_DATA/joint-crab/dl3/magic/run_05029748_DL3.fits")

    assert obs.rad_max is not None
    assert obs.rad_max.quantity.shape == (1, 1)
    assert u.allclose(obs.rad_max.quantity, 0.1414213 * u.deg)


@pytest.mark.xfail
@requires_data()
class TestObservationChecker:
    def setup_method(self):
        self.data_store = DataStore.from_dir("$GAMMAPY_DATA/cta-1dc/index/gps")

    def test_check_all(self):
        observation = self.data_store.obs(111140)
        records = list(observation.check())
        assert len(records) == 8

    def test_checker_bad(self):
        for index in range(5):
            self.data_store.hdu_table[index]["FILE_NAME"] = "bad"

        observation = self.data_store.obs(110380)
        records = list(observation.check())
        assert len(records) == 10
        assert records[1]["msg"] == "Loading events failed"
        assert records[3]["msg"] == "Loading GTI failed"
        assert records[5]["msg"] == "Loading aeff failed"
        assert records[7]["msg"] == "Loading edisp failed"
        assert records[9]["msg"] == "Loading psf failed"


@requires_data()
def test_observation_write(tmp_path):
    obs = Observation.read(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_023523.fits.gz"
    )
    mjdreff = obs.events.table.meta["MJDREFF"]
    mjdrefi = obs.events.table.meta["MJDREFI"]
    path = tmp_path / "obs.fits.gz"

    obs.meta.creation.origin = "test"
    obs.write(path)
    obs_read = obs.read(path)

    assert obs_read.events is not None
    assert obs_read.gti is not None
    assert obs_read.aeff is not None
    assert obs_read.edisp is not None
    assert obs_read.bkg is not None
    assert obs_read.rad_max is None
    assert obs_read.obs_id == 23523
    assert_allclose(obs_read.observatory_earth_location.lat.deg, -23.271778)

    assert_allclose(obs_read.events.table.meta["MJDREFF"], mjdreff)
    assert_allclose(obs_read.events.table.meta["MJDREFI"], mjdrefi)

    # unsupported format
    with pytest.raises(ValueError):
        obs.write(tmp_path / "foo.fits.gz", format="cool-new-format")

    # no irfs
    path = tmp_path / "obs_no_irfs.fits.gz"
    obs.write(path, include_irfs=False)
    obs_read = obs.read(path)
    assert obs_read.events is not None
    assert obs_read.gti is not None
    assert obs_read.aeff is None
    assert obs_read.edisp is None
    assert obs_read.bkg is None
    assert obs_read.rad_max is None


@requires_data()
def test_observation_read_write_checksum(tmp_path):
    obs = Observation.read(
        "$GAMMAPY_DATA/hess-dl3-dr1/data/hess_dl3_dr1_obs_id_023523.fits.gz"
    )
    path = tmp_path / "obs.fits"

    obs.write(path, checksum=True)

    with fits.open(path) as hdul:
        for hdu in hdul:
            assert "CHECKSUM" in hdu.header
            assert "DATASUM" in hdu.header

    with open(path, "r+b") as file:
        chunk = file.read(10000)
        index = chunk.find("EV_CLASS".encode("ascii"))
        file.seek(index)
        file.write("BAD__KEY".encode("ascii"))

    with pytest.warns(UserWarning):
        Observation.read(path, checksum=True)


@requires_data()
def test_obervation_copy(data_store):
    obs = data_store.obs(23523)

    obs_copy = obs.copy()
    assert obs_copy.obs_id == 23523
    assert isinstance(obs_copy.__dict__["_psf_hdu"], HDULocation)

    with pytest.raises(ValueError):
        _ = obs.copy(obs_id=1234)

    obs_copy = obs.copy(obs_id=1234, in_memory=True)
    assert isinstance(obs_copy.__dict__["psf"], PSF3D)
    assert obs_copy.obs_id == 1234


def test_observation_tmid():
    from gammapy.data import GTI

    start = Time("2020-01-01T20:00:00")
    stop = Time("2020-01-01T20:10:00")
    expected = Time("2020-01-01T20:05:00")
    epoch = Time("2020-01-01T00:00:00")

    gti = GTI.create([(start - epoch).to(u.s)], [(stop - epoch).to(u.s)], epoch)
    obs = Observation(gti=gti)
    assert abs(obs.tmid - expected).to(u.ns) < 1 * u.us


@requires_data()
def test_observations_clustering(data_store):
    selection = dict(
        type="sky_circle",
        frame="icrs",
        lon="83.633 deg",
        lat="22.014 deg",
        radius="2 deg",
    )
    obs_table = data_store.obs_table.select_observations(selection)
    observations = data_store.get_observations(obs_table["OBS_ID"])

    coord = SkyCoord(83.63308, 22.01450, unit="deg", frame="icrs")
    names = ["edisp-bias", "edisp-res", "psf-radius"]
    features = get_irfs_features(
        observations, energy_true="1 TeV", position=coord, names=names
    )

    n_features = len(names)
    features_array = np.array(
        [
            features[col].data
            for col in features.columns
            if col not in ["obs_id", "dataset_name"]
        ]
    ).T
    assert features_array.shape == (len(observations), n_features)

    features = hierarchical_clustering(
        features, linkage_kwargs={"method": "complete"}, fcluster_kwargs={"t": 2}
    )

    assert np.all(
        features["labels"].data
        == np.array(
            [
                1,
                1,
                2,
                2,
            ]
        )
    )

    features = get_irfs_features(
        observations,
        energy_true="1 TeV",
        position=coord,
        names=names,
        apply_standard_scaler=True,
    )
    features = hierarchical_clustering(features, fcluster_kwargs={"t": 2})
    features_array = np.array(
        [
            features[col].data
            for col in features.columns
            if col not in ["obs_id", "dataset_name"]
        ]
    ).T

    obs_clusters = observations.group_by_label(features["labels"])
    for k in range(n_features):
        assert_allclose(features_array[:, k].mean(), 0, atol=1e-7)
        assert_allclose(features_array[:, k].std(), 1, atol=1e-7)

    assert np.all(features["labels"].data == np.array([2, 1, 1, 1]))

    assert len(obs_clusters["group_1"]) == 3
    assert len(obs_clusters["group_2"]) == 1
    assert obs_clusters["group_2"][0].obs_id == 23523


@requires_data()
def test_filter_live_time_phase(data_store):
    observation = data_store.obs(20136)
    phase_filter = {"type": "custom", "opts": dict(parameter="PHASE", band=(0.2, 0.8))}

    default_obs_live_time = observation.observation_live_time_duration

    obs_filter = ObservationFilter(event_filters=[phase_filter])
    observation.obs_filter = obs_filter
    live_time_filter = observation.observation_live_time_duration

    assert_allclose(live_time_filter, default_obs_live_time * (0.8 - 0.2))


@requires_data()
def test_stack_observations(data_store, caplog):
    obs_1 = data_store.get_observations([20136, 20137, 20151])
    obs_2 = data_store.get_observations([20275, 20282])

    obs12 = Observations.from_stack([obs_1, obs_2])

    assert len(obs12) == 5
    assert isinstance(obs12[0], Observation)

    obs122 = Observations.from_stack([obs12, obs_2])

    assert len(obs122) == 7
    assert "WARNING" in [_.levelname for _ in caplog.records]
    assert "Observation with obs_id 20275 already belongs to Observations." in [
        _.message for _ in caplog.records
    ]

    caplog.clear()
    obs_1[2] = obs_1[0]

    assert "WARNING" in [_.levelname for _ in caplog.records]
    assert "Observation with obs_id 20136 already belongs to Observations." in [
        _.message for _ in caplog.records
    ]

    with pytest.raises(TypeError):
        Observations.from_stack([obs_1, ["a"]])


@requires_data()
def test_observations_generator(data_store):
    """Test Observations.generator()"""
    obs_1 = data_store.get_observations([20136, 20137, 20151])

    for idx, obs in enumerate(obs_1.in_memory_generator()):
        assert isinstance(obs, Observation)
        assert obs.obs_id == obs_1[idx].obs_id
        assert isinstance(obs.events, EventList)
        assert isinstance(obs.psf, PSF3D)


@requires_data()
def test_event_setter():
    irfs = load_irf_dict_from_file(
        "$GAMMAPY_DATA/cta-1dc/caldb/data/cta/1dc/bcf/South_z20_50h/irf_file.fits"
    )
    pointing = FixedPointingInfo(
        fixed_icrs=SkyCoord(0 * u.deg, 0 * u.deg),
    )
    location = EarthLocation(lon="-70d18m58.84s", lat="-24d41m0.34s", height="2000m")
    obs = Observation.create(
        obs_id=1,
        pointing=pointing,
        livetime=20 * u.min,
        irfs=irfs,
        location=location,
    )

    assert obs.events is None

    for invalid in (5, Table(), "foo"):
        with pytest.raises(TypeError):
            obs.events = invalid

    table = Table()
    table["RA"] = [0.0, 0.0, 0.0, 10.0] * u.deg
    table["DEC"] = [0.0, 0.9, 10.0, 10.0] * u.deg
    table["ENERGY"] = [1.0, 1.5, 1.5, 10.0] * u.TeV
    table["OFFSET"] = [0.1, 0.5, 1.0, 1.5] * u.deg
    table["TIME"] = Time("2025-01-01") + [0.1, 0.5, 1.0, 1.5] * u.s

    events = EventList(table)
    obs.events = events
    assert obs.events is events


@requires_data()
def test_observations_getitem(data_store):
    """Test mask indexing on Observations"""
    obs_1 = data_store.get_observations([20136, 20137, 20151])

    assert isinstance(obs_1[0], Observation)
    assert isinstance(obs_1[1:], Observations)
    assert len(obs_1[1:]) == 2

    mask = [True, False, True]
    obs_2 = obs_1[mask]

    assert len(obs_2) == 2
    assert obs_2.ids == ["20136", "20151"]

    ind = [0, 2]
    obs_2 = obs_1[ind]

    assert len(obs_2) == 2
    assert obs_2.ids == ["20136", "20151"]

    obs_2 = obs_1[np.array(mask)]

    assert len(obs_2) == 2
    assert obs_2.ids == ["20136", "20151"]

    assert obs_1[["20136", "20151"]].ids == ["20136", "20151"]

    obs_2 = obs_1[[]]
    assert len(obs_2) == 0
    assert isinstance(obs_2, Observations)

    obs_2 = obs_1[np.array([])]
    assert len(obs_2) == 0
    assert isinstance(obs_2, Observations)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy as np
from astropy.utils import minversion

__all__ = [
    "COPY_IF_NEEDED",
]

# This is copied from astropy.utils.compat
NUMPY_LT_2_0 = not minversion(np, "2.0.0.dev")

COPY_IF_NEEDED = False if NUMPY_LT_2_0 else None



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Gammapy integration and system tests.

This package can be used for tests that involve several
Gammapy sub-packages, or that don't fit anywhere else.
"""

import gammapy


def test_get_acknowledgment():
    text = gammapy.__acknowledgment__
    assert "Astropy" in text
    assert "iminuit" in text


def test_bibtex():
    text = gammapy.__bibtex__
    for item in ["article", "author", "DOI", "title", "journal"]:
        assert item in text
    assert "https://doi.org/10.1051/0004-6361/202346488" in text



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Utilities to compute J-factor maps."""
import html
import numpy as np
import astropy.units as u

__all__ = ["JFactory"]


class JFactory:
    """Compute J-Factor or D-Factor maps.

    J-Factors are computed for annihilation and D-Factors for decay.
    Set the argument `annihilation` to `False` to compute D-Factors.
    The assumed dark matter profiles will be centered on the center of the map.

    Parameters
    ----------
    geom : `~gammapy.maps.WcsGeom`
        Reference geometry.
    profile : `~gammapy.astro.darkmatter.profiles.DMProfile`
        Dark matter profile.
    distance : `~astropy.units.Quantity`
        Distance to convert angular scale of the map.
    annihilation: bool, optional
        Decay or annihilation. Default is True.
    """

    def __init__(self, geom, profile, distance, annihilation=True):
        self.geom = geom
        self.profile = profile
        self.distance = distance
        self.annihilation = annihilation

    def _repr_html_(self):
        try:
            return self.to_html()
        except AttributeError:
            return f"<pre>{html.escape(str(self))}</pre>"

    def compute_differential_jfactor(self, ndecade=1e4):
        r"""Compute differential J-Factor.

        .. math::
            \frac{\mathrm d J_\text{ann}}{\mathrm d \Omega} =
            \int_{\mathrm{LoS}} \mathrm d l \rho(l)^2

        .. math::
            \frac{\mathrm d J_\text{decay}}{\mathrm d \Omega} =
            \int_{\mathrm{LoS}} \mathrm d l \rho(l)
        """
        separation = self.geom.separation(self.geom.center_skydir).rad
        rmin = u.Quantity(
            value=np.tan(separation) * self.distance, unit=self.distance.unit
        )
        rmax = self.distance
        val = [
            (
                2
                * self.profile.integral(
                    _.value * u.kpc,
                    rmax,
                    np.arctan(_.value / self.distance.value),
                    ndecade,
                    self.annihilation,
                )
                + self.profile.integral(
                    self.distance,
                    4 * rmax,
                    np.arctan(_.value / self.distance.value),
                    ndecade,
                    self.annihilation,
                )
            )
            for _ in rmin.ravel()
        ]
        integral_unit = u.Unit("GeV2 cm-5") if self.annihilation else u.Unit("GeV cm-2")
        jfact = u.Quantity(val).to(integral_unit).reshape(rmin.shape)
        return jfact / u.steradian

    def compute_jfactor(self, ndecade=1e4):
        r"""Compute astrophysical J-Factor.

        .. math::
            J(\Delta\Omega) =
           \int_{\Delta\Omega} \mathrm d \Omega^{\prime}
           \frac{\mathrm d J}{\mathrm d \Omega^{\prime}}
        """
        diff_jfact = self.compute_differential_jfactor(ndecade)
        return diff_jfact * self.geom.to_image().solid_angle()
